from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd
from lamin_utils import colors, logger

if TYPE_CHECKING:
    from lamindb.base.types import FieldAttr, ListLike

    from .query_set import SQLRecordList
    from .sqlrecord import SQLRecord


# The base function for `from_values`
def _from_values(
    iterable: ListLike,
    field: FieldAttr,
    *,
    create: bool = False,
    organism: SQLRecord | str | None = None,
    source: SQLRecord | None = None,
    mute: bool = False,
) -> SQLRecordList:
    """Get or create records from iterables."""
    from .query_set import SQLRecordList

    registry = field.field.model  # type: ignore
    organism_record = get_organism_record_from_field(field, organism, values=iterable)
    # TODO: the create is problematic if field is not a name field
    if create:
        create_kwargs = {}
        if organism_record:
            create_kwargs["organism"] = organism_record
        return SQLRecordList(
            [
                registry(**{field.field.name: value}, **create_kwargs)
                for value in iterable
            ]
        )  # type: ignore

    iterable_idx = index_iterable(iterable)

    # returns existing records & non-existing values
    records, nonexist_values, msg = get_existing_records(
        iterable_idx=iterable_idx,
        field=field,
        organism=organism_record,
        mute=mute,
    )

    # new records to be created based on new values
    if len(nonexist_values) > 0:
        if registry.__base__.__name__ == "BioRecord":
            from bionty._organism import is_organism_required

            # if can and needed, get organism record from the existing records
            if (
                organism_record is None
                and len(records) > 0
                and is_organism_required(registry)
            ):
                organism_record = records[0].organism
            records_public, unmapped_values = create_records_from_source(
                iterable_idx=nonexist_values,
                field=field,
                organism=organism_record,
                source=source,
                msg=msg,
                mute=mute,
            )
            if len(records_public) > 0:
                msg = ""
            for record in records_public:
                record._from_source = True
            records += records_public
        else:
            unmapped_values = nonexist_values
        # unmapped new_ids will NOT create records
        if len(unmapped_values) > 0:
            # first log the success message
            if len(msg) > 0 and not mute:
                logger.success(msg)
            s = "" if len(unmapped_values) == 1 else "s"
            print_values = colors.yellow(_format_values(unmapped_values))
            n_nonval = colors.yellow(f"{len(unmapped_values)} non-validated")
            if not mute:
                logger.info(
                    f"{colors.red('did not create')} {registry.__name__} record{s} for "
                    f"{n_nonval} {colors.italic(f'{field.field.name}{s}')}: {print_values}"  # type: ignore
                )
    return SQLRecordList(records)


def get_existing_records(
    iterable_idx: pd.Index,
    field: FieldAttr,
    organism: SQLRecord | None = None,
    mute: bool = False,
) -> tuple[list, pd.Index, str]:
    """Get existing records from the database."""
    # NOTE: existing records matching is agnostic to the source
    model = field.field.model  # type: ignore

    # log synonyms mapped terms
    syn_mapper = model.standardize(
        iterable_idx,
        field=field,
        organism=organism,
        mute=True,
        source_aware=False,  # standardize only based on the DB reference
        return_mapper=True,
    )
    iterable_idx = iterable_idx.to_frame().rename(index=syn_mapper).index

    # now we have to sort the list of queried records
    # preserved = Case(
    #     *[
    #         When(**{field.field.name: value}, then=pos)
    #         for pos, value in enumerate(iterable_idx)
    #     ]
    # )
    # order by causes a factor 10 in runtime
    # records = query_set.order_by(preserved).to_list()

    # log validated terms
    is_validated = model.validate(
        iterable_idx, field=field, organism=organism, mute=True
    )
    if len(is_validated) > 0:
        validated = iterable_idx[is_validated]
    else:
        validated = []
    msg = ""
    syn_msg = ""
    if not mute:
        if len(validated) > 0:
            s = "" if len(validated) == 1 else "s"
            print_values = colors.green(_format_values(validated))
            msg = (
                "loaded"
                f" {colors.green(f'{len(validated)} {model.__name__} record{s}')}"
                f" matching {colors.italic(f'{field.field.name}')}: {print_values}"
            )
        if len(syn_mapper) > 0:
            s = "" if len(syn_mapper) == 1 else "s"
            names = list(syn_mapper.keys())
            print_values = colors.green(_format_values(names))
            syn_msg = (
                "loaded"
                f" {colors.green(f'{len(syn_mapper)} {model.__name__} record{s}')}"
                f" matching {colors.italic('synonyms')}: {print_values}"
            )

    # no logging if all values are validated
    # logs if there are synonyms
    if len(syn_msg) > 0:
        if len(msg) > 0 and not mute:
            logger.success(msg)
        if not mute:
            logger.success(syn_msg)
        msg = ""

    # get all existing records in the db
    query = {f"{field.field.name}__in": iterable_idx.values}  # type: ignore
    if organism is not None:
        query["organism"] = organism
    records = model.filter(**query).to_list()

    if len(validated) == len(iterable_idx):
        return records, pd.Index([]), msg
    else:
        nonval_values = iterable_idx.difference(validated)
        return records, nonval_values, msg


def create_records_from_source(
    iterable_idx: pd.Index,
    field: FieldAttr,
    organism: SQLRecord | None = None,
    source: SQLRecord | None = None,
    msg: str = "",
    mute: bool = False,
) -> tuple[list, pd.Index]:
    """Create records from source."""
    model = field.field.model  # type: ignore
    records: list = []
    # populate additional fields from public_df
    from bionty._source import filter_public_df_columns, get_source_record

    # get the default source
    source_record = get_source_record(model, organism, source)

    # create the corresponding PublicOntology object from model
    try:
        public_ontology = model.public(source=source_record)
    except Exception:
        # no public source
        return records, iterable_idx

    # filter the columns in public df based on fields
    public_df = filter_public_df_columns(model=model, public_ontology=public_ontology)

    if public_df.empty:
        return records, iterable_idx

    # standardize in the public reference
    # do not inspect synonyms if the field is not name field
    inspect_synonyms = True
    if hasattr(model, "_name_field") and field.field.name != model._name_field:  # type: ignore
        inspect_synonyms = False
    result = public_ontology.inspect(
        iterable_idx,
        field=field.field.name,  # type: ignore
        mute=True,
        inspect_synonyms=inspect_synonyms,
    )
    syn_mapper = result.synonyms_mapper

    msg_syn: str = ""
    if len(syn_mapper) > 0:
        s = "" if len(syn_mapper) == 1 else "s"
        names = list(syn_mapper.keys())
        print_values = colors.purple(_format_values(names))
        msg_syn = (
            "created"
            f" {colors.purple(f'{len(syn_mapper)} {model.__name__} record{s} from Bionty')}"
            f" matching {colors.italic('synonyms')}: {print_values}"
        )

        iterable_idx = iterable_idx.to_frame().rename(index=syn_mapper).index

    # create records for values that are found in the public reference
    # matching either field or synonyms
    mapped_values = iterable_idx.intersection(public_df[field.field.name])  # type: ignore

    multi_msg = ""
    if len(mapped_values) > 0:
        public_kwargs, multi_msg = _bulk_create_dicts_from_df(
            keys=mapped_values,
            column_name=field.field.name,  # type: ignore
            df=public_df,
        )

        # this here is needed when the organism is required to create new records
        if organism is None:
            organism = get_organism_record_from_field(
                field, source_record.organism, values=mapped_values
            )

        create_kwargs = (
            {"organism": organism, "source": source_record}
            if organism is not None
            else {"source": source_record}
        )
        for bk in public_kwargs:
            records.append(model(**bk, **create_kwargs, _skip_validation=True))

        # number of records that matches field (not synonyms)
        validated = result.validated
        if len(validated) > 0:
            s = "" if len(validated) == 1 else "s"
            print_values = colors.purple(_format_values(validated))
            # this is the success msg for existing records in the DB
            if len(msg) > 0 and not mute:
                logger.success(msg)
            if not mute:
                logger.success(
                    "created"
                    f" {colors.purple(f'{len(validated)} {model.__name__} record{s} from Bionty')}"
                    f" matching {colors.italic(f'{field.field.name}')}: {print_values}"  # type: ignore
                )

    # make sure that synonyms logging appears after the field logging
    if len(msg_syn) > 0 and not mute:
        logger.success(msg_syn)
    # warning about multi matches
    if len(multi_msg) > 0 and not mute:
        logger.warning(multi_msg)

    # return the values that are not found in the public reference
    unmapped_values = iterable_idx.difference(mapped_values)
    return records, unmapped_values


def index_iterable(iterable: ListLike) -> pd.Index:
    """Get unique values from an iterable."""
    idx = pd.Index(iterable).unique()
    # No entries are made for NAs, '', None
    # returns an ordered unique not null list
    return idx[(idx != "") & (~idx.isnull())]


def _format_values(
    names: ListLike, n: int = 20, quotes: bool = True, sep: str = "'"
) -> str:
    """Format values for printing."""
    if isinstance(names, dict):
        items = {
            f"{key}: {value}": None
            for key, value in names.items()
            if key != "None" and value != "None"
        }
    else:
        # Use a dictionary instead of a list to have unique values and preserve order
        items = {str(name): None for name in names if name != "None"}

    unique_items = list(items.keys())

    if quotes:
        unique_items = [f"{sep}{item}{sep}" for item in unique_items]

    print_values = ", ".join(unique_items[:n])

    if len(unique_items) > n:
        print_values += ", ..."

    return print_values


def _bulk_create_dicts_from_df(
    keys: set | list, column_name: str, df: pd.DataFrame
) -> tuple[dict, str]:
    """Get fields from a DataFrame for many rows."""
    multi_msg = ""
    if df.index.name != column_name:
        df = df.set_index(column_name).loc[list(keys)]
    if not df.index.is_unique:
        # return all records for multi-matches with a warning
        dup = df.index[df.index.duplicated()].unique().tolist()
        if len(dup) > 0:
            s = "" if len(dup) == 1 else "s"
            print_values = _format_values(dup)
            multi_msg = (
                f"ambiguous validation in Bionty for {len(dup)} record{s}:"
                f" {print_values}"
            )

    return df.reset_index().to_dict(orient="records"), multi_msg


def get_organism_record_from_field(  # type: ignore
    field: FieldAttr,
    organism: str | SQLRecord | None = None,
    values: ListLike = None,
    using_key: str | None = None,
) -> SQLRecord | None:
    """Get organism record.

    Args:
        field: the field to get the organism record for
        organism: the organism to get the record for
        values: the values to get the organism record for
        using_key: the db to get the organism record for

    Returns:
        The organism record if:
            The organism FK is required for the registry
            The field is not unique or the organism is not None
    """
    if values is None:
        values = []
    registry = field.field.model
    field_str = field.field.name
    # id field is a unique field that's not a relation
    is_simple_field_unique = field.field.unique and not field.field.is_relation
    check = not is_simple_field_unique or organism is not None

    if (
        registry.__get_name_with_module__() == "bionty.Gene"
        and field.field.name == "ensembl_gene_id"
        and len(values) > 0
        and organism is None
    ):  # type: ignore
        from bionty._organism import organism_from_ensembl_id

        return organism_from_ensembl_id(values[0], using_key)  # type: ignore

    if registry.__base__.__name__ == "BioRecord" and check:
        from bionty._organism import create_or_get_organism_record

        organism_record = create_or_get_organism_record(
            organism=organism, registry=registry, field=field_str
        )
        return organism_record



from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING

import lamindb_setup as ln_setup
from lamin_utils import colors, logger
from lamindb_setup import settings as setup_settings
from lamindb_setup._set_managed_storage import set_managed_storage
from lamindb_setup.core import deprecated
from lamindb_setup.core._settings_instance import sanitize_git_repo_url
from lamindb_setup.core._settings_storage import StorageSettings

from .subsettings._annotation_settings import AnnotationSettings, annotation_settings
from .subsettings._creation_settings import CreationSettings, creation_settings

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from upath import UPath


def is_read_only_connection() -> bool:
    instance = setup_settings.instance
    if instance.dialect == "postgresql":
        db_url = instance.db
        return "read" in db_url or "public" in db_url
    return False


VERBOSITY_TO_INT = {
    "error": 0,  # 40
    "warning": 1,  # 30
    "success": 2,  # 25
    "info": 3,  # 20
    "hint": 4,  # 15
    "debug": 5,  # 10
}
VERBOSITY_TO_STR: dict[int, str] = dict(
    [reversed(i) for i in VERBOSITY_TO_INT.items()]  # type: ignore
)


class Settings:
    """Settings.

    Please use the global `ln.settings` object instead of instantiating this class yourself.
    """

    def __init__(self):
        self._verbosity_int: int = 1  # warning-level logging
        logger.set_verbosity(self._verbosity_int)
        self._sync_git_repo: str | None = None

    def __repr__(self) -> str:  # pragma: no cover
        if "sphinx" in sys.modules:
            return object.__repr__(self)

        cls_name = colors.green(self.__class__.__name__)
        verbosity_color = colors.yellow if self.verbosity == "warning" else colors.green
        verbosity_str = verbosity_color(self.verbosity)

        storage_root = self._storage_settings.root_as_str
        storage_str = colors.italic(storage_root)

        instance_str = colors.italic(self.instance_uid)
        track_color = colors.green if self.track_run_inputs else colors.yellow
        track_str = track_color(str(self.track_run_inputs))

        lines = [
            f"{cls_name}",
            f"  instance: {instance_str}",
            f"  storage: {storage_str}",
            f"  verbosity: {verbosity_str}",
            f"  track_run_inputs: {track_str}",
        ]

        if self.sync_git_repo:
            repo_name = (
                self.sync_git_repo.split("/")[-1]
                if "/" in self.sync_git_repo
                else self.sync_git_repo
            )
            lines.append(f"  sync_git_repo: {colors.italic(repo_name)}")

        return "\n".join(lines)

    @property
    def creation(self) -> CreationSettings:
        """SQLRecord creation settings.

        For example, `ln.settings.creation.search_names = False` will disable
        searching for records with similar names during creation.
        """
        return creation_settings

    @property
    def annotation(self) -> AnnotationSettings:
        """Artifact annotation settings.

        For example, `ln.settings.creation.search_names = False` will disable
        searching for records with similar names during creation.
        """
        return annotation_settings

    # note: this setting should probably be deprecated soon
    # warnings could then be filtered with a regular warning mechanism
    track_run_inputs: bool = True
    """Track run inputs (default `True`).

    If this setting is true, an artifact is recorded as run input upon `.load()`, `.cache()` & `.open()` provided :func:`~lamindb.track` was called in the current compute (Python, R) session.
    If :func:`~lamindb.track` was not called, you receive a warning message upon `.load()`, `.cache()` & `.open()`.

    If you switch this setting to `False`, you won't see the warning message anymore and no run inputs will be recorded.

    FAQ: :doc:`/faq/track-run-inputs`
    """
    __using_key: str | None = None
    _using_storage: str | None = None

    @property
    def _using_key(self) -> str | None:
        """Key for Django database settings."""
        return self.__using_key

    @_using_key.setter
    def _using_key(self, value: str | None):
        ln_setup.settings._using_key = value
        self.__using_key = value

    @property
    def _storage_settings(self) -> ln_setup.core.StorageSettings:
        if self._using_storage is None:
            storage_settings = ln_setup.settings.storage
        else:
            storage_settings = ln_setup.core.StorageSettings(root=self._using_storage)
        return storage_settings

    @property
    def sync_git_repo(self) -> str | None:
        """Sync transforms with scripts in git repository.

        Provide the full git repo URL.
        """
        if self._sync_git_repo is not None:
            return self._sync_git_repo
        elif os.environ.get("LAMINDB_MULTI_INSTANCE") == "true":
            return None
        else:
            return setup_settings.instance.git_repo

    @sync_git_repo.setter
    def sync_git_repo(self, value) -> None:
        """Sync transforms with scripts in git repository.

        For example: `ln.settings.sync_git_repo = https://github.com/laminlabs/redun-lamin`
        """
        self._sync_git_repo = sanitize_git_repo_url(value)
        if not self._sync_git_repo.startswith("https://"):  # pragma: nocover
            raise ValueError("git repository URL must start with 'https://'.")

    @property
    def storage(self) -> StorageSettings:
        """Current default storage location for writes.

        Examples:

        Retrieve the storage settings::

            ln.settings.storage
            #> StorageSettings(root='s3://my-bucket')

        Retrieve the storage root::

            ln.settings.storage.root
            #> UPath('s3://my-bucket')

        You can write artifacts to other storage locations by switching the current default storage location::

            ln.settings.storage = "s3://some-bucket"

        You can also pass additional fsspec kwargs via::

            kwargs = dict(
                profile="some_profile", # fsspec arg
                cache_regions=True # fsspec arg for s3
            )
            ln.settings.storage = "s3://some-bucket", kwargs
        """
        return self._storage_settings

    @storage.setter
    def storage(self, path_kwargs: str | Path | UPath | tuple[str | UPath, Mapping]):
        import lamindb as ln

        if isinstance(path_kwargs, tuple):
            path, kwargs = path_kwargs
            # we should ultimately deprecate passing host here, I think
            if isinstance(kwargs, str):
                kwargs = {"host": kwargs}
        else:
            path, kwargs = path_kwargs, {}
        ssettings = StorageSettings(root=path)  # there is no need to pass kwargs here!
        exists = ln.Storage.filter(root=ssettings.root_as_str).one_or_none()
        if exists is None:
            response = input(
                f"Storage location {ssettings.root_as_str} does not yet exist. Do you want to continue with creating it? (y/n) "
            )
            # logger.warning(f"deprecated call because storage location does **not yet** exist; going forward, please create through ln.Storage(root={path}).save() going forward")
            if response != "y":
                return None
            set_managed_storage(path, **kwargs)
        else:
            if exists.instance_uid != ln_setup.settings.instance.uid:
                raise ValueError(
                    f"Storage {ssettings.root_as_str} exists in another instance ({exists.instance_uid}), cannot write to it from here."
                )
            ssettings = StorageSettings(
                root=exists.root,
                region=exists.region,
                uid=exists.uid,
                instance_id=ln_setup.settings.instance._id,
            )
            ln_setup.settings.instance._storage = ssettings
            kwargs.pop("host", None)  # host is not needed for existing storage
            settings.storage._set_fs_kwargs(**kwargs)

    @property
    def instance_uid(self) -> str:
        """The `uid` of the current instance."""
        return ln_setup.settings.instance.uid

    @property
    def cache_dir(self) -> UPath:
        """Cache root, a local directory to cache cloud files."""
        return ln_setup.settings.cache_dir

    @property
    def local_storage(self) -> StorageSettings:
        """An additional local default storage (a path to its root).

        Is only available if :attr:`~lamindb.setup.core.InstanceSettings.keep_artifacts_local` is enabled.

        Guide: :doc:`faq/keep-artifacts-local`
        """
        return ln_setup.settings.instance.local_storage

    @local_storage.setter
    def local_storage(self, local_root: Path):
        import lamindb as ln

        # note duplication with storage setter!
        ssettings = StorageSettings(root=local_root)
        exists = ln.Storage.filter(root=ssettings.root_as_str).one_or_none()
        if exists is None:
            response = input(
                f"Storage location {ssettings.root_as_str} does not yet exist. Do you want to continue with creating it? (y/n) "
            )
            # logger.warning(f"deprecated call because storage location does **not yet** exist; going forward, please create through ln.Storage(root={path}).save() going forward")
            if response != "y":
                return None
        else:
            if exists.instance_uid != ln_setup.settings.instance.uid:
                raise ValueError(
                    f"Storage {ssettings.root_as_str} exists in another instance ({exists.instance_uid}), cannot write to it from here."
                )
        ln_setup.settings.instance.local_storage = local_root

    @property
    @deprecated("local_storage")
    def storage_local(self) -> StorageSettings:
        return self.local_storage

    @storage_local.setter
    @deprecated("local_storage")
    def storage_local(self, local_root_host: tuple[Path | str, str]):
        self.local_storage = local_root_host  # type: ignore

    @property
    def verbosity(self) -> str:
        """Logger verbosity (default `'warning'`).

        - `'error'`: only show error messages
        - `'warning'`: also show warning messages
        - `'success'`: also show success and save messages
        - `'info'`: also show info messages
        - `'hint'`: also show hint messages
        - `'debug'`: also show detailed debug messages
        """
        return VERBOSITY_TO_STR[self._verbosity_int]

    @verbosity.setter
    def verbosity(self, verbosity: str | int):
        if isinstance(verbosity, str):
            verbosity_int = VERBOSITY_TO_INT[verbosity]
        else:
            verbosity_int = verbosity
        self._verbosity_int = verbosity_int
        logger.set_verbosity(verbosity_int)


settings = Settings()



import lamindb_setup as _lamindb_setup
from lamindb_setup import *  # noqa: F403
from lamindb_setup import (
    connect,
    delete,
    init,
    settings,
)

from . import core, errors, types
from ._switch import switch  # noqa: F401

del connect  # we have this at the root level, hence, we don't want it here
__doc__ = _lamindb_setup.__doc__.replace("lamindb_setup", "lamindb.setup")
settings.__doc__ = settings.__doc__.replace("lamindb_setup", "lamindb.setup")



from pathlib import Path

import lamindb as ln
import pytest
from lamindb_setup.core._hub_core import get_storage_records_for_instance


def check_storage_location_on_hub_exists(uid: str):
    all_storage_records = get_storage_records_for_instance(
        ln.setup.settings.instance._id
    )
    length = len([r for r in all_storage_records if r["lnid"] == uid])
    if length not in {0, 1}:
        raise AssertionError(
            f"Expected 0 or 1 storage records for uid {uid}, found {length}."
        )
    return length == 1


def test_reference_storage_location(ccaplog):
    ln.Artifact("s3://lamindata/iris_studies/study0_raw_images")
    assert ln.Storage.get(root="s3://lamindata").instance_uid == "4XIuR0tvaiXM"
    # assert (
    #     "referenced read-only storage location at s3://lamindata, is managed by instance with uid 4XIuR0tvaiXM"
    #     in ccaplog.text
    # )


def test_switch_delete_storage_location():
    ln.settings.storage = "./default_storage_unit_storage"
    assert (
        ln.settings.storage.root.resolve()
        == Path("./default_storage_unit_storage").resolve()
    )
    new_storage_location = "s3://lamindb-ci/test-settings-switch-storage"
    ln.Storage(new_storage_location).save()
    ln.settings.storage = new_storage_location
    assert ln.setup.settings.storage.type_is_cloud
    assert ln.setup.settings.storage.root_as_str == new_storage_location
    # root.fs contains the underlying fsspec filesystem
    # the following is set by lamindb to True for s3 by default
    assert ln.setup.settings.storage.root.fs.cache_regions
    ln.settings.storage = new_storage_location, {"cache_regions": False}
    assert not ln.setup.settings.storage.root.fs.cache_regions
    assert ln.setup.settings.storage.root.exists()

    # now work with the new storage location
    new_storage = ln.Storage.get(root=new_storage_location)
    assert check_storage_location_on_hub_exists(new_storage.uid)
    artifact = ln.Artifact(".gitignore", key="test_artifact").save()
    assert new_storage.root in artifact.path.as_posix()

    # artifacts exist
    with pytest.raises(AssertionError) as err:
        new_storage.delete()
    assert "Cannot delete storage holding artifacts." in err.exconly()

    artifact.delete(permanent=True, storage=False)
    # still some files in there
    with pytest.raises(ln.setup.errors.StorageNotEmpty) as err:
        new_storage.delete()
    assert (
        "'s3://lamindb-ci/test-settings-switch-storage/.lamindb' contains 1 objects"
        in err.exconly()
    )

    # now delete the artifact so that the storage location is empty
    artifact.path.unlink()
    with pytest.raises(AssertionError) as err:
        new_storage.delete()
    assert (
        "Cannot delete the current storage location, switch to another."
        in err.exconly()
    )

    # check all attempts unsuccessful so far
    assert check_storage_location_on_hub_exists(new_storage.uid)

    # switch back to default storage
    ln.settings.storage = "./default_storage_unit_storage"
    storage_marker = ln.UPath(new_storage_location) / ".lamindb/storage_uid.txt"
    assert storage_marker.exists()
    new_storage.delete()
    assert not check_storage_location_on_hub_exists(new_storage.uid)
    assert not storage_marker.exists()



from __future__ import annotations

import subprocess
import sys
from typing import TYPE_CHECKING

import lamindb_setup as ln_setup
from lamin_utils import logger

if TYPE_CHECKING:
    from lamindb.models import Run


def track_python_environment(run: Run) -> None:
    env_dir = ln_setup.settings.cache_dir / "environments" / f"run_{run.uid}"
    filepath = env_dir / "run_env_pip.txt"
    if not env_dir.exists():
        filepath.parent.mkdir(parents=True)
    # create a requirements.txt
    # we don't create a conda environment.yml mostly for its slowness
    try:
        with open(filepath, "w") as f:
            result = subprocess.run(
                [sys.executable, "-m", "pip", "freeze"],
                stdout=f,
            )
    except OSError as e:
        result = None
        logger.warning(f"could not run pip freeze with error {e}")
    if result is not None and result.returncode == 0:
        logger.info(f"tracked pip freeze > {str(filepath)}")



from __future__ import annotations

from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Literal

from django.db.models import Case, Q, TextField, Value, When
from django.db.models.functions import Concat
from lamin_utils import logger
from lamindb_setup.core._docs import doc_args
from upath import UPath

from ..core._mapped_collection import MappedCollection
from ..core.storage._backed_access import _open_dataframe
from .artifact import Artifact, track_run_input
from .collection import Collection, _load_concat_artifacts

if TYPE_CHECKING:
    from anndata import AnnData
    from lamindb_setup.types import UPathStr
    from pandas import DataFrame
    from polars import LazyFrame as PolarsLazyFrame
    from pyarrow.dataset import Dataset as PyArrowDataset


UNORDERED_WARNING = (
    "this query set is unordered, consider using `.order_by()` first "
    "to avoid opening the artifacts in an arbitrary order"
)


# maybe make this abstract
class ArtifactSet(Iterable):
    """Abstract class representing sets of artifacts returned by queries.

    This class automatically extends :class:`~lamindb.models.BasicQuerySet`
    and :class:`~lamindb.models.QuerySet` when the base model is :class:`~lamindb.Artifact`.

    Examples:

        >>> artifacts = ln.Artifact.filter(otype="AnnData")
        >>> artifacts # an instance of ArtifactQuerySet inheriting from ArtifactSet
    """

    @doc_args(Collection.load.__doc__)
    def load(
        self,
        join: Literal["inner", "outer"] = "outer",
        is_run_input: bool | None = None,
        **kwargs,
    ) -> DataFrame | AnnData:
        """{}"""  # noqa: D415
        if not self.ordered:  # type: ignore
            logger.warning(UNORDERED_WARNING)

        artifacts: list[Artifact] = list(self)
        concat_object = _load_concat_artifacts(artifacts, join, **kwargs)
        # track only if successful
        track_run_input(artifacts, is_run_input)
        return concat_object

    @doc_args(Collection.open.__doc__)
    def open(
        self,
        engine: Literal["pyarrow", "polars"] = "pyarrow",
        is_run_input: bool | None = None,
        **kwargs,
    ) -> PyArrowDataset | Iterator[PolarsLazyFrame]:
        """{}"""  # noqa: D415
        if not self.ordered:  # type: ignore
            logger.warning(UNORDERED_WARNING)

        artifacts: list[Artifact] = list(self)
        paths: list[UPath] = [artifact.path for artifact in artifacts]

        dataframe = _open_dataframe(paths, engine=engine, **kwargs)
        # track only if successful
        track_run_input(artifacts, is_run_input)
        return dataframe

    @doc_args(Collection.mapped.__doc__)
    def mapped(
        self,
        layers_keys: str | list[str] | None = None,
        obs_keys: str | list[str] | None = None,
        obsm_keys: str | list[str] | None = None,
        obs_filter: dict[str, str | list[str]] | None = None,
        join: Literal["inner", "outer"] | None = "inner",
        encode_labels: bool | list[str] = True,
        unknown_label: str | dict[str, str] | None = None,
        cache_categories: bool = True,
        parallel: bool = False,
        dtype: str | None = None,
        stream: bool = False,
        is_run_input: bool | None = None,
    ) -> MappedCollection:
        """{}"""  # noqa: D415
        if not self.ordered:  # type: ignore
            logger.warning(UNORDERED_WARNING)

        artifacts: list[Artifact] = []
        paths: list[UPath] = []
        for artifact in self:
            if ".h5ad" not in artifact.suffix and ".zarr" not in artifact.suffix:
                logger.warning(f"ignoring artifact with suffix {artifact.suffix}")
                continue
            elif not stream:
                paths.append(artifact.cache())
            else:
                paths.append(artifact.path)
            artifacts.append(artifact)
        ds = MappedCollection(
            paths,
            layers_keys,
            obs_keys,
            obsm_keys,
            obs_filter,
            join,
            encode_labels,
            unknown_label,
            cache_categories,
            parallel,
            dtype,
        )
        # track only if successful
        track_run_input(artifacts, is_run_input)
        return ds


def artifacts_from_path(artifacts: ArtifactSet, path: UPathStr) -> ArtifactSet:
    """Returns artifacts in the query set that are registered for the provided path."""
    from lamindb.models import BasicQuerySet, QuerySet

    # not QuerySet but only BasicQuerySet
    assert isinstance(artifacts, BasicQuerySet) and not isinstance(artifacts, QuerySet)  # noqa: S101

    upath = UPath(path)

    path_str = upath.as_posix()

    stem = upath.stem
    stem_len = len(stem)

    if stem_len == 16:
        qs = artifacts.filter(
            Q(_key_is_virtual=True) | Q(key__isnull=True),
            _real_key__isnull=True,
            uid__startswith=stem,
        )
    elif stem_len == 20:
        qs = artifacts.filter(
            Q(_key_is_virtual=True) | Q(key__isnull=True),
            _real_key__isnull=True,
            uid=stem,
        )
    else:
        qs = None

    if qs:  # an empty query set evaluates to False
        return qs

    qs = (
        artifacts.filter(Q(_key_is_virtual=False) | Q(_real_key__isnull=False))
        .alias(
            db_path=Case(
                When(
                    _real_key__isnull=False,
                    then=Concat(
                        "storage__root",
                        Value("/"),
                        "_real_key",
                        output_field=TextField(),
                    ),
                ),
                default=Concat(
                    "storage__root", Value("/"), "key", output_field=TextField()
                ),
                output_field=TextField(),
            )
        )
        .filter(db_path=path_str)
    )

    return qs



import bionty as bt
import lamindb as ln
import pytest
from lamindb.errors import ValidationError


# some validate tests are in test_queryset
def test_inspect():
    ln.Schema.filter().delete(permanent=True)
    bt.Gene.filter().delete(permanent=True)
    bt.settings.organism = "human"
    result = bt.Gene.inspect("TCF7", "symbol")
    assert result.validated == []

    bt.Gene.from_source(symbol="TCF7").save()
    result = bt.Gene.inspect("TCF7")
    assert bt.Gene.validate("TCF7", organism="human")
    result = bt.Gene.inspect(["TCF7", "ABC1"], "symbol")
    assert result.validated == ["TCF7"]

    # clean up
    bt.Gene.filter().delete(permanent=True)


# if a record was added to the DB via a different source
# it will still be validated because it's in the DB
def test_inspect_source():
    source1 = bt.Source.get(entity="bionty.CellType", name="cl")
    source2 = bt.CellType.add_source(source="cl", version="2022-08-16")
    bt.CellType.from_source(name="T cell", source=source1).save()
    assert bt.CellType.inspect("T-cell", source=source2, mute=True).synonyms_mapper == {
        "T-cell": "T cell"
    }
    assert (
        bt.CellType.inspect(
            "T-cell", source=source2, mute=True, strict_source=True
        ).synonyms_mapper
        == {}
    )
    assert bt.CellType.validate("T cell", source=source2, mute=True).sum() == 1
    assert (
        bt.CellType.validate(
            "T cell", source=source2, mute=True, strict_source=True
        ).sum()
        == 0
    )
    assert bt.CellType.standardize("T-cell", source=source2, mute=True) == "T cell"
    # here still standardized because of bionty
    assert (
        bt.CellType.standardize("T-cell", source=source2, mute=True, strict_source=True)
        == "T cell"
    )
    bt.CellType.filter().delete(permanent=True)


def test_standardize():
    bt.settings.organism = "human"

    # synonym not in the database
    result = bt.Gene.standardize(["ABC1", "PDCD1"])
    assert result == ["HEATR6", "PDCD1"]

    result = bt.Gene.standardize(["ABC1", "PDCD1"], field=bt.Gene.symbol)
    assert result == ["HEATR6", "PDCD1"]

    mapper = bt.Gene.standardize(["ABC1", "PDCD1"], return_mapper=True)
    assert mapper == {"ABC1": "HEATR6"}

    # synonym already in the database
    bt.Gene.from_source(symbol="LMNA").save()
    mapper = bt.Gene.standardize(["ABC1", "LMN1"], return_mapper=True)
    assert mapper == {"LMN1": "LMNA", "ABC1": "HEATR6"}
    assert bt.Gene.standardize(["LMNA"]) == ["LMNA"]
    assert bt.Gene.standardize("LMNA") == "LMNA"
    assert bt.Gene.standardize(["LMN1"], return_mapper=True) == {"LMN1": "LMNA"}


def test_standardize_source_aware():
    result = bt.Gene.standardize(["ABC1", "PDCD1"], source_aware=False)
    assert result == ["ABC1", "PDCD1"]


def test_add_remove_synonym():
    bt.CellType.filter().delete(permanent=True)

    # a registry that doesn't have a synonyms column
    user = ln.User.get(handle="testuser1")
    with pytest.raises(NotImplementedError):
        user.add_synonym("syn")

    cell_types = bt.CellType.from_values(["T cell", "B cell"], "name")
    ln.save(cell_types)
    tcell = bt.CellType.get(name="T cell")
    bcell = bt.CellType.get(name="B cell")
    tcell.add_synonym(["my cell type"])
    tcell.add_synonym("")
    tcell.add_synonym([])
    assert "my cell type" in tcell.synonyms
    with pytest.raises(ValidationError):
        bcell.add_synonym("my cell type")
    with pytest.raises(ValidationError):
        tcell.add_synonym("my|celltype")

    tcell.remove_synonym("my cell type")
    assert "my cell type" not in tcell.synonyms

    bcell.synonyms = None
    bcell.save()
    tcell.synonyms = None
    tcell.save()
    tcell.add_synonym("")
    tcell.add_synonym([""])
    tcell.add_synonym([])
    tcell.add_synonym(["my cell type"])
    tcell.add_synonym("")
    tcell.add_synonym([""])
    tcell.add_synonym([])
    assert tcell.synonyms == "my cell type"
    tcell.remove_synonym("my cell type")

    # clean up
    bt.CellType.filter().delete(permanent=True)


def test_set_abbr():
    bt.CellType.filter().delete(permanent=True)
    bt.CellType(name="my cell type").save()
    record = bt.CellType.get(name="my cell type")
    # if abbr is name, do not add to synonyms
    record.set_abbr("my cell type")
    assert record.abbr == "my cell type"
    assert record.synonyms is None

    record.set_abbr("myct")
    assert record.abbr == "myct"
    assert "myct" in record.synonyms

    source = bt.Source.filter(organism="human").first()
    with pytest.raises(AttributeError) as error:
        source.set_abbr("abbr")
    assert (
        error.exconly() == "AttributeError: 'Source' object has no attribute 'set_abbr'"
    )

    record.delete()


def test_validate_int():
    result = ln.User.validate([1, 2], field=ln.User.id)
    assert result.sum() == 1


def test_synonym_mapping():
    # only name field can be standardized
    bt.Gene.from_source(symbol="TNFRSF4", organism="human").save()

    result = bt.Gene.inspect(
        ["CD134", "TNFRSF4"], field=bt.Gene.symbol, organism="human"
    )
    assert result.synonyms_mapper == {"CD134": "TNFRSF4"}

    result = bt.Gene.inspect(
        ["CD134", "TNFRSF4"], field=bt.Gene.ensembl_gene_id, organism="human"
    )
    assert result.synonyms_mapper == {}

    bt.Gene.filter().delete(permanent=True)



from __future__ import annotations

import builtins
import re
from datetime import datetime, timezone
from time import sleep
from typing import TYPE_CHECKING

import lamindb_setup as ln_setup
from lamin_utils import logger
from lamin_utils._logger import LEVEL_TO_COLORS, LEVEL_TO_ICONS, RESET_COLOR
from lamindb_setup.core.hashing import hash_dir, hash_file

from lamindb.models import Artifact, Run, Transform

is_run_from_ipython = getattr(builtins, "__IPYTHON__", False)

if TYPE_CHECKING:
    from pathlib import Path


def get_save_notebook_message() -> str:
    # do not add bold() or any other complicated characters as then we can't match this
    # easily anymore in an html to strip it out
    return f"please hit {get_shortcut()} to save the notebook in your editor"


def get_save_notebook_message_retry() -> str:
    return f"{get_save_notebook_message()} and re-run finish()"


# this code was originally in nbproject by the same authors
def check_consecutiveness(
    nb, calling_statement: str = None, silent_success: bool = True
) -> bool:
    """Check whether code cells have been executed consecutively.

    Needs to be called in the last code cell of a notebook.
    Otherwise raises `RuntimeError`.

    Returns cell transitions that violate execution at increments of 1 as a list
    of tuples.

    Args:
        nb: Notebook content.
        calling_statement: The statement that calls this function.
    """
    cells = nb.cells

    violations = []
    prev = 0

    ccount = 0  # need to initialize because notebook might note have code cells
    # and below, we check if ccount is None
    for cell in cells:
        cell_source = "".join(cell["source"])
        if cell["cell_type"] != "code" or cell_source == "":
            continue

        if calling_statement is not None and calling_statement in cell_source:
            continue

        ccount = cell["execution_count"]
        if ccount is None or prev is None or ccount - prev != 1:
            violations.append((prev, ccount))

        prev = ccount

    # ignore the very last code cell of the notebook
    # `check_consecutiveness` is being run during publish if `last_cell`` is True
    # hence, that cell has ccount is None
    if ccount is None:
        violations.pop()

    any_violations = len(violations) > 0
    if any_violations:
        logger.warning(f"cells {violations} were not run consecutively")
    elif not silent_success:
        logger.success("cell execution numbers increase consecutively")

    return not any_violations


def get_shortcut() -> str:
    import platform

    return "CMD + s" if platform.system() == "Darwin" else "CTRL + s"


def get_seconds_since_modified(filepath) -> float:
    return datetime.now().timestamp() - filepath.stat().st_mtime


def save_run_logs(run: Run, save_run: bool = False) -> None:
    logs_path = ln_setup.settings.cache_dir / f"run_logs_{run.uid}.txt"
    if logs_path.exists():
        if run.report is not None:
            logger.important("overwriting run.report")
        artifact = Artifact(  # type: ignore
            logs_path,
            description=f"log streams of run {run.uid}",
            kind="__lamindb_run__",
            run=False,
        )
        artifact.save(upload=True, print_progress=False)
        run.report = artifact
        if save_run:  # defaults to false because is slow
            run.save()


# this is from the get_title function in nbproject
# should be moved into lamindb sooner or later
def prepare_notebook(
    nb,
    strip_title: bool = False,
) -> str | None:
    title_found = False
    for cell in nb.cells:
        cell.metadata.clear()  # strip cell metadata
        if not title_found and cell["cell_type"] == "markdown":
            lines = cell["source"].split("\n")
            for i, line in enumerate(lines):
                if line.startswith("# "):
                    line.lstrip("#").strip(" .").strip()
                    title_found = True
                    if strip_title:
                        lines.pop(i)
                        cell["source"] = "\n".join(lines)
        # strip logging message about saving notebook in editor
        # this is normally the last cell
        if cell["cell_type"] == "code" and ".finish(" in cell["source"]:
            for output in cell["outputs"]:
                if "to save the notebook in your editor" in output.get("text", ""):
                    cell["outputs"] = []
                    break
    return None


def notebook_to_report(notebook_path: Path, output_path: Path) -> None:
    import nbformat
    import traitlets.config as config
    from nbconvert import HTMLExporter

    with open(notebook_path, encoding="utf-8") as f:
        notebook = nbformat.read(f, as_version=4)
    prepare_notebook(notebook, strip_title=True)
    notebook.metadata.clear()  # strip notebook metadata
    # if we were to export as ipynb, the following two lines would do it
    # with open(output_path, "w", encoding="utf-8") as f:
    #     nbformat.write(notebook, f)
    # instead we need all this code
    c = config.Config()
    c.HTMLExporter.preprocessors = []
    c.HTMLExporter.exclude_input_prompt = True
    c.HTMLExporter.exclude_output_prompt = True
    c.HTMLExporter.anchor_link_text = " "
    html_exporter = HTMLExporter(config=c)
    html, _ = html_exporter.from_notebook_node(notebook)
    output_path.write_text(html, encoding="utf-8")


def notebook_to_script(  # type: ignore
    title: str, notebook_path: Path, script_path: Path | None = None
) -> None | str:
    import jupytext

    notebook = jupytext.read(notebook_path)
    py_content = jupytext.writes(notebook, fmt="py:percent")
    # remove global metadata header
    py_content = re.sub(r"^# ---\n.*?# ---\n\n", "", py_content, flags=re.DOTALL)
    # replace title
    py_content = py_content.replace(f"# # {title}", "#")
    if script_path is None:
        return py_content
    else:
        script_path.write_text(py_content, encoding="utf-8")


def clean_r_notebook_html(file_path: Path) -> tuple[str | None, Path]:
    import re

    cleaned_content = file_path.read_text()
    # remove title from content
    pattern_title = r"<title>(.*?)</title>"
    title_match = re.search(pattern_title, cleaned_content)
    title_text = None
    if title_match:
        title_text = title_match.group(1)
        pattern_h1 = f"<h1[^>]*>{re.escape(title_text)}</h1>"
        cleaned_content = re.sub(pattern_title, "", cleaned_content)
        cleaned_content = re.sub(pattern_h1, "", cleaned_content)
    # remove error message from content
    if "to save the notebook in your editor" in cleaned_content:
        orig_error_message = f"! {get_save_notebook_message_retry()}"
        # coming up with the regex for this is a bit tricky due to all the
        # escape characters we'd need to insert into the message; hence,
        # we do this with a replace() instead
        cleaned_content = cleaned_content.replace(orig_error_message, "")
        if "to save the notebook in your editor" in cleaned_content:
            orig_error_message = orig_error_message.replace(
                " finish()", "\nfinish()"
            )  # RStudio might insert a newline
            cleaned_content = cleaned_content.replace(orig_error_message, "")
    cleaned_path = file_path.parent / (f"{file_path.stem}.cleaned{file_path.suffix}")
    cleaned_path.write_text(cleaned_content, encoding="utf-8")
    return title_text, cleaned_path


def check_filepath_recently_saved(filepath: Path, is_finish_retry: bool) -> bool:
    # the recently_saved_time needs to be very low for the first check
    # because an accidental save (e.g. via auto-save) might otherwise lead
    # to upload of an outdated notebook
    # also see implementation for R notebooks below
    offset_saved_time = 0.3 if not is_finish_retry else 20
    for retry in range(30):
        recently_saved_time = offset_saved_time + retry  # sleep time is 1 sec
        if get_seconds_since_modified(filepath) > recently_saved_time:
            if retry == 0:
                prefix = f"{LEVEL_TO_COLORS[20]}{LEVEL_TO_ICONS[20]}{RESET_COLOR}"
                print(f"{prefix} {get_save_notebook_message()}", end=" ")
            elif retry == 9:
                print(".", end="\n")
            elif retry == 4:
                print(". still waiting ", end="")
            else:
                print(".", end="")
            sleep(1)
        else:
            if retry > 0:
                prefix = f"{LEVEL_TO_COLORS[25]}{LEVEL_TO_ICONS[25]}{RESET_COLOR}"
                print(f" {prefix}")
            # filepath was recently saved, return True
            return True
    # if we arrive here, no save event occured, return False
    return False


def save_context_core(
    *,
    run: Run | None,
    transform: Transform,
    filepath: Path,
    finished_at: bool = False,
    skip_save_report: bool = False,
    ignore_non_consecutive: bool | None = None,
    from_cli: bool = False,
    is_retry: bool = False,
    notebook_runner: str | None = None,
) -> str | None:
    import lamindb as ln
    from lamindb.models import (
        format_field_value,  # needs to come after lamindb was imported because of CLI use
    )

    ln.settings.verbosity = "success"

    # for scripts, things are easy
    is_consecutive = True
    is_ipynb = filepath.suffix == ".ipynb"
    is_r_notebook = filepath.suffix in {".qmd", ".Rmd"}
    source_code_path = filepath
    report_path: Path | None = None
    save_source_code_and_report = filepath.exists()
    if (
        is_run_from_ipython and notebook_runner != "nbconvert" and filepath.exists()
    ):  # python notebooks in interactive session
        if is_ipynb:
            # ignore this for py:percent notebooks
            import nbproject

            # it might be that the user modifies the title just before ln.finish()
            if (nbproject_title := nbproject.meta.live.title) != transform.description:
                transform.description = nbproject_title
                transform.save()
        if not ln_setup._TESTING:
            save_source_code_and_report = check_filepath_recently_saved(
                filepath, is_retry
            )
            if not save_source_code_and_report and not is_retry:
                logger.warning(get_save_notebook_message_retry())
                return "retry"
            elif not save_source_code_and_report:
                logger.warning(
                    "the notebook on disk wasn't saved within the last 10 sec"
                )
    if is_ipynb and filepath.exists():  # could be from CLI outside interactive session
        try:
            import jupytext  # noqa: F401
            from nbproject.dev import (
                read_notebook,
            )
        except ImportError:
            logger.error("install nbproject & jupytext: pip install nbproject jupytext")
            return None
        notebook_content = read_notebook(filepath)  # type: ignore
        if not ignore_non_consecutive:  # ignore_non_consecutive is None or False
            is_consecutive = check_consecutiveness(
                notebook_content, calling_statement=".finish("
            )
            if not is_consecutive:
                response = "n"  # ignore_non_consecutive == False
                if ignore_non_consecutive is None:  # only print warning
                    response = "y"  # we already printed the warning
                else:  # ask user to confirm
                    response = input(
                        "   Do you still want to proceed with finishing? (y/n) "
                    )
                if response != "y":
                    return "aborted-non-consecutive"
        # write the report
        report_path = ln_setup.settings.cache_dir / filepath.name.replace(
            ".ipynb", ".html"
        )
        notebook_to_report(filepath, report_path)
        # write the source code
        source_code_path = ln_setup.settings.cache_dir / filepath.name.replace(
            ".ipynb", ".py"
        )
        notebook_to_script(transform.description, filepath, source_code_path)
    elif is_ipynb and not filepath.exists():
        logger.warning("notebook file does not exist in compute environment")
    elif is_r_notebook:
        if filepath.with_suffix(".nb.html").exists():
            report_path = filepath.with_suffix(".nb.html")
        elif filepath.with_suffix(".html").exists():
            report_path = filepath.with_suffix(".html")
        else:
            logger.warning(
                f"no html report found; to attach one, create an .html export for your {filepath.suffix} file and then run: lamin save {filepath}"
            )
    if report_path is not None and is_r_notebook and not from_cli:  # R notebooks
        # see comment above in check_filepath_recently_saved
        recently_saved_time = 0.3 if not is_retry else 20
        if get_seconds_since_modified(report_path) > recently_saved_time:
            # the automated retry solution of Jupyter notebooks does not work in RStudio because the execution of the notebook cell
            # seems to block the event loop of the frontend
            if not is_retry:
                logger.warning(get_save_notebook_message_retry())
                return "retry"
            else:
                logger.warning(
                    "the notebook on disk hasn't been saved within the last 20 sec"
                )
            save_source_code_and_report = False
    ln.settings.creation.artifact_silence_missing_run_warning = True
    # save source code
    if save_source_code_and_report:
        transform_hash, _ = hash_file(source_code_path)  # ignore hash_type for now
        if transform.hash is not None:
            # check if the hash of the transform source code matches
            if transform_hash != transform.hash:
                response = input(
                    f"You are about to overwrite existing source code (hash '{transform.hash}') for Transform('{transform.uid}')."
                    f" Proceed? (y/n) "
                )
                if response == "y":
                    transform.source_code = source_code_path.read_text()
                    transform.hash = transform_hash
                else:
                    logger.warning("Please re-run `ln.track()` to make a new version")
                    return "rerun-the-notebook"
            else:
                logger.debug("source code is already saved")
        else:
            transform.source_code = source_code_path.read_text()
            transform.hash = transform_hash

    if run is not None:
        base_path = ln_setup.settings.cache_dir / "environments" / f"run_{run.uid}"
        paths = [base_path / "run_env_pip.txt", base_path / "r_environment.txt"]
        existing_paths = [path for path in paths if path.exists()]
        if len(existing_paths) == 2:
            # let's not store the python environment for an R session for now
            existing_paths = [base_path / "r_environment.txt"]

        if existing_paths:
            overwrite_env = True
            if run.environment_id is not None and from_cli:
                logger.important("run.environment is already saved, ignoring")
                overwrite_env = False

            if overwrite_env:
                # Use directory if multiple files exist, otherwise use the single file
                artifact_path: Path = (
                    base_path if len(existing_paths) > 1 else existing_paths[0]
                )

                # Set description based on what we're saving
                if len(existing_paths) == 1:
                    if existing_paths[0].name == "run_env_pip.txt":
                        description = "requirements.txt"
                    elif existing_paths[0].name == "r_environment.txt":
                        description = "r_environment.txt"
                    env_hash, _ = hash_file(artifact_path)
                else:
                    description = "environments"
                    _, env_hash, _, _ = hash_dir(artifact_path)

                artifact = ln.Artifact.objects.filter(hash=env_hash).one_or_none()
                new_env_artifact = artifact is None

                if new_env_artifact:
                    artifact = ln.Artifact(
                        artifact_path,
                        description=description,
                        kind="__lamindb_run__",
                        run=False,
                    )
                    artifact.save(upload=True, print_progress=False)

                run.environment = artifact
                if new_env_artifact:
                    logger.debug(f"saved run.environment: {run.environment}")

    # set finished_at
    if finished_at and run is not None:
        if not from_cli:
            update_finished_at = True
        else:
            update_finished_at = run.finished_at is None
        if update_finished_at:
            run.finished_at = datetime.now(timezone.utc)

    # track report and set is_consecutive
    if save_source_code_and_report and not skip_save_report:
        if run is not None:
            # do not save a run report if executing through nbconvert
            if report_path is not None and notebook_runner != "nbconvert":
                if is_r_notebook:
                    title_text, report_path = clean_r_notebook_html(report_path)
                    if title_text is not None:
                        transform.description = title_text
                if run.report_id is not None:
                    hash, _ = hash_file(report_path)  # ignore hash_type for now
                    if hash != run.report.hash:
                        response = input(
                            f"You are about to overwrite an existing report (hash '{run.report.hash}') for Run('{run.uid}'). Proceed? (y/n) "
                        )
                        if response == "y":
                            run.report.replace(report_path)
                            run.report.save(upload=True, print_progress=False)
                        else:
                            logger.important("keeping old report")
                    else:
                        logger.important("report is already saved")
                else:
                    report_file = ln.Artifact(  # type: ignore
                        report_path,
                        description=f"Report of run {run.uid}",
                        kind="__lamindb_run__",  # hidden file
                        run=False,
                    )
                    report_file.save(upload=True, print_progress=False)
                    run.report = report_file
                if is_r_notebook:
                    # this is the "cleaned" report
                    report_path.unlink()
                logger.debug(
                    f"saved transform.latest_run.report: {transform.latest_run.report}"
                )
            run._is_consecutive = is_consecutive
        if report_path is not None and notebook_runner == "nbconvert":
            logger.important(f"to save the notebook html, run: lamin save {filepath}")

    # save both run & transform records if we arrive here
    if run is not None:
        run.save()
    transform_id_prior_to_save = transform.id
    transform.save()  # this in-place updates the state of transform upon hash collision
    if transform.id != transform_id_prior_to_save:
        # the hash existed and we're actually back to the previous version
        # hence, this was in fact a run of the previous transform rather than of
        # the new transform
        # this can happen in interactively executed notebooks with a pro-active version bump in case it turns out that the user didn't make a change to the notebook
        run.transform = transform
        run.save()
        ln.Transform.get(transform_id_prior_to_save).delete(permanent=True)

    # finalize
    if finished_at and not from_cli and run is not None:
        run_time = run.finished_at - run.started_at
        days = run_time.days
        seconds = run_time.seconds
        hours = seconds // 3600
        minutes = (seconds % 3600) // 60
        secs = seconds % 60
        formatted_run_time = (
            f"{days}d"
            if days != 0
            else "" + f"{hours}h"
            if hours != 0
            else "" + f"{minutes}m"
            if minutes != 0
            else "" + f"{secs}s"
        )

        logger.important(
            f"finished Run('{run.uid}') after {formatted_run_time} at {format_field_value(run.finished_at)}"
        )
    if ln_setup.settings.instance.is_on_hub:
        instance_slug = ln_setup.settings.instance.slug
        if save_source_code_and_report:
            ui_url = ln_setup.settings.instance.ui_url
            logger.important(
                f"go to: {ui_url}/{instance_slug}/transform/{transform.uid}"
            )
        if finished_at and not from_cli and save_source_code_and_report:
            thing = "notebook" if (is_ipynb or is_r_notebook) else "script"
            logger.important(
                f"to update your {thing} from the CLI, run: lamin save {filepath}"
            )
    if not save_source_code_and_report:
        logger.warning(
            f"did *not* save source code and report -- to do so, run: lamin save {filepath}"
        )
    return None



"""Examples and utilities for Mlflow.

.. autosummary::
   :toctree: .

   save_mlflow_features
"""

import lamindb as ln


def save_mlflow_features():
    """Saves all MLflow experiment and run related features.

    Saves the following features:

    - mlflow_run_id
    - mlflow_run_name
    - mlflow_experiment_id
    - mlflow_experiment_name
    - mlflow_user_id
    - mlflow_status
    - mlflow_lifecycle_stage
    - mlflow_artifact_uri
    - mlflow_start_time
    - mlflow_end_time
    """
    mlflow_type = ln.Feature(name="MLflow", is_type=True).save()
    ln.Feature(name="mlflow_run_id", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_run_name", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_experiment_id", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_experiment_name", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_user_id", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_status", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_lifecycle_stage", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_artifact_uri", dtype=str, type=mlflow_type).save()
    ln.Feature(name="mlflow_start_time", dtype=int, type=mlflow_type).save()
    ln.Feature(name="mlflow_end_time", dtype=int, type=mlflow_type).save()



# Note: Almost all logic for schema-based validation is handled in the curators test suite
# This here only covers external feature annotation and validation
from pathlib import Path

import lamindb as ln
import pandas as pd
import pytest


@pytest.fixture(scope="module")
def two_internal_features():
    feat1 = ln.Feature(name="feat1", dtype=int).save()
    feat2 = ln.Feature(name="feat2", dtype=int).save()
    yield feat1, feat2
    feat1.delete(permanent=True)
    feat2.delete(permanent=True)


@pytest.fixture(scope="module")
def two_external_features():
    featA = ln.Feature(name="species", dtype=str).save()
    featB = ln.Feature(name="split", dtype=str).save()
    yield featA, featB
    featA.delete(permanent=True)
    featB.delete(permanent=True)


@pytest.mark.parametrize("use_schema", [True, False])
def test_create_artifact_with_external_feature_annotations(
    tsv_file: Path,
    use_schema: bool,
    two_external_features: tuple[ln.Feature, ln.Feature],
):
    feat1, feat2 = two_external_features
    if use_schema:
        schema = ln.Schema(features=[feat1, feat2]).save()
    else:
        schema = None
    artifact = ln.Artifact(
        tsv_file,
        key="test.tsv",
        features={"species": "bird", "split": "train"},
        schema=schema,
    ).save()
    assert artifact.features.get_values() == {"species": "bird", "split": "train"}
    artifact.delete(permanent=True)
    if use_schema:
        schema.delete(permanent=True)


def test_from_dataframe_with_external_features_and_schema(
    df: pd.DataFrame,
    two_external_features: tuple[ln.Feature, ln.Feature],
    two_internal_features: tuple[ln.Feature, ln.Feature],
):
    feat1, feat2 = two_internal_features
    featA, featB = two_external_features
    schema_external = ln.Schema(features=[featA, featB]).save()

    # Case 1: wrong internal features for this dataframe
    schema_with_mistake = ln.Schema(
        features=[featA, featB],
        slots={"__external__": schema_external},
        otype="DataFrame",
    ).save()
    with pytest.raises(ln.errors.ValidationError) as error:
        artifact = ln.Artifact.from_dataframe(
            df,
            key="test_df_with_external_features.parquet",
            features={"species": "bird", "split": "train"},
            schema=schema_with_mistake,
        ).save()
    assert "COLUMN_NOT_IN_DATAFRAME" in error.exconly()

    # Case 2: no schema for external features provided
    schema_no_external = ln.Schema(features=[feat1, feat2]).save()
    artifact = ln.Artifact.from_dataframe(
        df,
        key="test_df_with_external_features.parquet",
        features={"species": "bird", "split": "train"},
        schema=schema_no_external,
    ).save()
    assert artifact.features.get_values() == {"species": "bird", "split": "train"}
    artifact.delete(permanent=True)

    # Case 3: correct external schema
    schema_correct_external = ln.Schema(
        features=[feat1, feat2],
        slots={"__external__": schema_external},
        otype="DataFrame",
    ).save()
    artifact = ln.Artifact.from_dataframe(
        df,
        key="test_df_with_external_features.parquet",
        features={"species": "bird", "split": "train"},
        schema=schema_correct_external,
    ).save()
    assert artifact.features.get_values() == {"species": "bird", "split": "train"}

    inferred_schema = artifact.feature_sets.all()[0]
    artifact.feature_sets.remove(inferred_schema.id)
    inferred_schema.delete(permanent=True)
    artifact.delete(permanent=True)
    schema_with_mistake.delete(permanent=True)
    schema_no_external.delete(permanent=True)
    schema_correct_external.delete(permanent=True)
    schema_external.delete(permanent=True)



import bionty as bt
import lamindb as ln
import pandas as pd
import pytest
from lamindb.errors import ValidationError
from lamindb.models.feature import serialize_pandas_dtype
from pandas.api.types import is_string_dtype


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        {
            "feat1": [1, 2, 3],
            "feat2": [3.1, 4.2, 5.3],
            "feat3": ["cond1", "cond2", "cond2"],
            "feat4": ["id1", "id2", "id3"],
            "rando_feature": ["rando1", "rando2", "rando3"],
        }
    )


@pytest.fixture(scope="module")
def dict_data():
    return {
        "dict_feat1": 42,
        "dict_feat2": 3.14,
        "dict_feat3": "somestring",  # string (ambiguous cat ? str)
        "dict_feat4": True,
        "dict_feat5": [1, 2, 3],
        "dict_feat6": ["a", "b", "c"],  # list[str] (ambiguous list[cat ? str])
        "dict_feat7": {"key": "value"},
    }


def test_feature_init():
    # positional args not supported
    with pytest.raises(ValueError):
        ln.Feature("x")

    # dtype required unless is_type=True
    with pytest.raises(ValidationError):
        ln.Feature(name="feat")

    # is OK if also is_type is passed
    ln.Feature(name="Feat", is_type=True)

    # invalid dtype string
    with pytest.raises(ValueError):
        ln.Feature(name="feat", dtype="x")

    # categorical dtype must specify valid types
    with pytest.raises(ValidationError):
        ln.Feature(name="feat", dtype="cat[1]")

    # ensure feat1 does not exist
    if feat1 := ln.Feature.filter(name="feat1").one_or_none() is not None:
        feat1.delete(permanent=True)

    feat1 = ln.Feature(name="feat", dtype="str").save()
    # duplicate name with different dtype should fail
    with pytest.raises(ValidationError) as error:
        ln.Feature(name="feat", dtype="cat")
    assert (
        error.exconly()
        == "lamindb.errors.ValidationError: Feature feat already exists with dtype str, you passed cat"
    )
    feat1.delete(permanent=True)

    # string and list syntax for categorical dtypes should be equivalent and work
    feat2 = ln.Feature(name="feat2", dtype="str", description="feat2").save()
    feat2_again = ln.Feature(name="feat2", dtype="str", description="feat2").save()
    assert feat2 == feat2_again
    feat2.delete(permanent=True)

    # categorical dtype with union of registries using string syntax must be valid
    feature = ln.Feature(name="feat1", dtype="cat[Record|bionty.Gene]")
    assert feature.dtype == "cat[Record|bionty.Gene]"
    # categorical dtype with union of registries using objects must be valid
    feature = ln.Feature(name="feat1", dtype=[ln.Record, bt.Gene])
    assert feature.dtype == "cat[Record|bionty.Gene]"

    # dtype with field name before bracket filters must be valid
    feature = ln.Feature(
        name="gene_feature", dtype="cat[bionty.Gene.ensembl_gene_id[organism='human']]"
    )
    assert "bionty.Gene" in feature.dtype
    assert "ensembl_gene_id" in feature.dtype
    assert "organism='human'" in feature.dtype


def test_cat_filters_dtype():
    feature = ln.Feature(
        name="disease",
        dtype=bt.Disease,
        cat_filters={
            "source__uid": "4a3ejKuf"
        },  # uid corresponds to disease_ontology_old.uid
    ).save()

    assert feature.dtype == "cat[bionty.Disease[source__uid='4a3ejKuf']]"

    feature.delete(permanent=True)


def test_cat_filters_empty_filter():
    # empty filter values should be rejected
    with pytest.raises(ValidationError) as error:
        ln.Feature(name="feat_empty", dtype=bt.Disease, cat_filters={"source__uid": ""})
    assert (
        "lamindb.errors.ValidationError: Empty value in filter source__uid"
        in error.exconly()
    )


def test_cat_filters_invalid_field_name():
    # invalid filter field names should be rejected
    source = bt.Source(
        name="", description="", organism="", entity="", version=""
    ).save()
    with pytest.raises(ValidationError) as error:
        ln.Feature(
            name="feat_invalid_attr",
            dtype=bt.Disease,
            cat_filters={"source__invalid_field": source},
        )
    assert (
        "lamindb.errors.ValidationError: SQLRecord Source has no attribute 'invalid_field' in filter source__invalid_field"
        in error.exconly()
    )
    source.delete(permanent=True)


def test_feature_from_df(df):
    if feat1 := ln.Feature.filter(name="feat1").one_or_none() is not None:
        feat1.delete(permanent=True)
    features = ln.Feature.from_dataframe(df.iloc[:, :4]).save()
    artifact = ln.Artifact.from_dataframe(df, description="test").save()
    # test for deprecated add_feature_set
    artifact.features._add_schema(ln.Schema(features), slot="columns")
    features = artifact.features.slots["columns"].features.all()
    assert len(features) == len(df.columns[:4])
    [col for col in df.columns if is_string_dtype(df[col])]
    categoricals = {
        col: df[col] for col in df.columns if isinstance(df[col], pd.CategoricalDtype)
    }
    for feature in features:
        if feature.name in categoricals:
            assert feature.dtype == "cat"
        else:
            orig_type = df[feature.name].dtype
            assert feature.dtype == serialize_pandas_dtype(orig_type)
    for feature in features:
        feature.save()
    labels = [ln.Record(name=name) for name in df["feat3"].unique()]
    ln.save(labels)
    feature = ln.Feature.get(name="feat3")
    feature.dtype = "cat"
    feature.save()
    with pytest.raises(ValidationError) as err:
        artifact.labels.add(labels, feature=feature)
    assert (
        err.exconly()
        == "lamindb.errors.ValidationError: Cannot manually annotate a feature measured *within* the dataset. Please use a Curator."
    )
    extfeature = ln.Feature(name="extfeat", dtype="str").save()
    with pytest.raises(ValidationError) as err:
        artifact.labels.add(labels, feature=extfeature)
    assert (
        err.exconly()
        == f"lamindb.errors.ValidationError: Feature {extfeature.name} needs dtype='cat' for label annotation, currently has dtype='str'"
    )

    # clean up
    artifact.delete(permanent=True)
    ln.Schema.filter().delete(permanent=True)
    ln.Record.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)


def test_feature_from_dict(dict_data):
    # defaults to str for ambiguous types
    features = ln.Feature.from_dict(dict_data)
    assert len(features) == len(dict_data)
    assert features[0].dtype == "int"
    assert features[1].dtype == "float"
    assert features[2].dtype == "str"
    assert features[3].dtype == "bool"
    assert features[4].dtype == "list[int]"
    assert features[5].dtype == "list[str]"
    assert features[6].dtype == "dict"

    # deprecated: convert str to cat
    with pytest.warns(DeprecationWarning, match="str_as_cat.*deprecated"):
        features = ln.Feature.from_dict(dict_data, str_as_cat=True)
    assert len(features) == len(dict_data)
    assert features[2].dtype == "cat"
    assert features[5].dtype == "list[cat]"

    # deprecated: do not convert str to cat
    with pytest.warns(DeprecationWarning, match="str_as_cat.*deprecated"):
        features = ln.Feature.from_dict(dict_data, str_as_cat=False)
    assert features[2].dtype == "str"
    assert features[5].dtype == "list[str]"

    # Wrong field
    with pytest.raises(ValueError) as e:
        ln.Feature.from_dict(dict_data, field=ln.Record.name)
    assert "field must be a Feature FieldAttr" in str(e.value)

    # Explicit field
    features_with_field = ln.Feature.from_dict(dict_data, field=ln.Feature.name)
    assert len(features_with_field) == len(dict_data)


def test_feature_from_dict_type(dict_data):
    feature_type = ln.Feature(name="Testdata_feature_type", is_type=True).save()
    with pytest.warns(DeprecationWarning, match="str_as_cat.*deprecated"):
        features = ln.Feature.from_dict(
            dict_data, str_as_cat=True, type=feature_type
        ).save()
    for feature in features:
        assert feature.type.name == "Testdata_feature_type"
    ln.Feature.filter(type__isnull=False).delete(permanent=True)
    feature_type.delete(permanent=True)



# ruff: noqa: TC004
from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from datetime import date, datetime
from itertools import compress
from typing import TYPE_CHECKING, Any, MutableMapping

import anndata as ad
import numpy as np
import pandas as pd
from anndata import AnnData
from django.contrib.postgres.aggregates import ArrayAgg
from django.db import connections
from django.db.models import Aggregate, ProtectedError, Subquery
from django.db.utils import IntegrityError
from lamin_utils import logger
from lamindb_setup.core.hashing import hash_set
from lamindb_setup.core.upath import create_path
from lamindb_setup.errors import ModuleWasntConfigured
from rich.table import Column, Table
from rich.text import Text

from lamindb.core.storage import LocalPathClasses
from lamindb.errors import DoesNotExist, InvalidArgument, ValidationError
from lamindb.models._from_values import _format_values
from lamindb.models.feature import (
    serialize_pandas_dtype,
    suggest_categorical_for_str_iterable,
)
from lamindb.models.save import save
from lamindb.models.schema import DICT_KEYS_TYPE, Schema
from lamindb.models.sqlrecord import (
    REGISTRY_UNIQUE_FIELD,
    get_name_field,
    transfer_fk_to_default_db_bulk,
    transfer_to_default_db,
)

from ..base import deprecated
from ._describe import (
    NAME_WIDTH,
    TYPE_WIDTH,
    VALUES_WIDTH,
    describe_header,
    format_rich_tree,
)
from ._django import get_artifact_with_related
from ._label_manager import _get_labels, describe_labels
from ._relations import (
    dict_related_model_to_related_name,
)
from .feature import Feature, FeatureValue, parse_dtype
from .sqlrecord import SQLRecord
from .ulabel import ULabel

if TYPE_CHECKING:
    from rich.tree import Tree

    from lamindb.base.types import FieldAttr
    from lamindb.models import (
        Artifact,
        Collection,
        IsLink,
    )
    from lamindb.models.query_set import BasicQuerySet

    from .run import Run


def get_accessor_by_registry_(host: Artifact | Collection) -> dict:
    dictionary = {
        field.related_model.__get_name_with_module__(): field.name
        for field in host._meta.related_objects
    }
    dictionary["Feature"] = "features"
    dictionary["ULabel"] = "ulabels"
    dictionary["Record"] = "records"
    return dictionary


def get_schema_by_slot_(host: Artifact) -> dict[str, Schema]:
    # if the host is not yet saved
    if host._state.adding:
        if hasattr(host, "_staged_feature_sets"):
            return host._staged_feature_sets
        else:
            return {}
    host_db = host._state.db
    kwargs = {"artifact_id": host.id}
    # otherwise, we need a query
    links_schema = (
        host.feature_sets.through.objects.using(host_db)
        .filter(**kwargs)
        .select_related("schema")
    )
    return {fsl.slot: fsl.schema for fsl in links_schema}


def get_label_links(
    host: Artifact | Collection, registry: str, feature: Feature
) -> BasicQuerySet:
    kwargs = {"artifact_id": host.id, "feature_id": feature.id}
    link_records = (
        getattr(host, host.features._accessor_by_registry[registry])  # type: ignore
        .through.objects.using(host._state.db)
        .filter(**kwargs)
    )
    return link_records


def get_schema_links(host: Artifact | Collection) -> BasicQuerySet:
    kwargs = {"artifact_id": host.id}
    links_schema = host.feature_sets.through.objects.filter(**kwargs)
    return links_schema


def get_link_attr(link: IsLink | type[IsLink], data: Artifact | Collection) -> str:
    link_model_name = link.__class__.__name__
    if link_model_name in {"Registry", "ModelBase"}:  # we passed the type of the link
        link_model_name = link.__name__  # type: ignore
    return link_model_name.replace(data.__class__.__name__, "").lower()


# Custom aggregation for SQLite
class GroupConcat(Aggregate):
    function = "GROUP_CONCAT"
    template = '%(function)s(%(expressions)s, ", ")'


def custom_aggregate(field, using: str):
    if connections[using].vendor == "postgresql":
        return ArrayAgg(field)
    else:
        return GroupConcat(field)


def _get_categoricals_postgres(
    self: Artifact | Collection | Run,
    related_data: dict | None = None,
) -> dict[tuple[str, str], set[str]]:
    """Get categorical features and their values using PostgreSQL-specific optimizations."""
    if not related_data:
        if self.__class__.__name__ == "Artifact":
            artifact_meta = get_artifact_with_related(
                self, include_feature_link=True, include_m2m=True
            )
            related_data = artifact_meta.get("related_data", {})
        else:
            related_data = {}

    # Process m2m data
    m2m_data = related_data.get("m2m", {}) if related_data else {}
    m2m_name = {}
    for related_name, values in m2m_data.items():
        link_model = getattr(self.__class__, related_name).through
        related_model_name = link_model.__name__.replace(
            self.__class__.__name__, ""
        ).lower()
        m2m_name[related_model_name] = values

    # Get feature information
    links_data = related_data.get("link", {}) if related_data else {}
    feature_dict = {
        id: (name, dtype)
        for id, name, dtype in Feature.objects.using(self._state.db).values_list(
            "id", "name", "dtype"
        )
    }

    # Build result dictionary
    result = defaultdict(set)
    for link_name, link_values in links_data.items():
        related_name = link_name.removeprefix("links_").replace("_", "")
        if not link_values:
            continue

        for link_value in link_values:
            feature_id = link_value.get("feature")
            if feature_id is None:
                continue

            feature_name, feature_dtype = feature_dict.get(feature_id)
            label_id = link_value.get(related_name)
            label_name = m2m_name.get(related_name, {}).get(label_id)
            if label_name:
                result[(feature_name, feature_dtype)].add(label_name)

    return dict(result)


def _get_categoricals(
    self: Artifact | Collection,
) -> dict[tuple[str, str], set[str]]:
    """Get categorical features and their values using the default approach."""
    result = defaultdict(set)
    for _, links in _get_labels(self, links=True, instance=self._state.db).items():
        for link in links:
            if hasattr(link, "feature_id") and link.feature_id is not None:
                feature = Feature.objects.using(self._state.db).get(id=link.feature_id)
                link_attr = get_link_attr(link, self)
                label = getattr(link, link_attr)
                name_attr = (
                    "name" if hasattr(label, "name") else label.__class__._name_field
                )
                label_name = getattr(label, name_attr)
                result[(feature.name, feature.dtype)].add(label_name)

    return dict(result)


def _get_non_categoricals(
    self,
) -> dict[tuple[str, str], set[Any]]:
    """Get non-categorical features and their values."""
    from .artifact import Artifact
    from .run import Run

    non_categoricals = {}

    if self.id is not None and isinstance(self, (Artifact, Run)):
        attr_name = "feature"
        _feature_values = (
            getattr(self, f"_{attr_name}_values")
            .values(f"{attr_name}__name", f"{attr_name}__dtype")
            .annotate(values=custom_aggregate("value", self._state.db))
            .order_by(f"{attr_name}__name")
        )

        for fv in _feature_values:
            feature_name = fv[f"{attr_name}__name"]
            feature_dtype = fv[f"{attr_name}__dtype"]
            values = fv["values"]

            # Convert single values to sets
            if not isinstance(values, (list, dict, set)):
                values = {values}
            elif (
                isinstance(values, list)
                and feature_dtype != "dict"
                and not feature_dtype.startswith("list")
            ):
                try:
                    values = set(values)
                except TypeError:
                    # TypeError: unhashable type: 'list' if values is list[list]
                    pass

            # Handle special datetime types
            if feature_dtype == "datetime":
                values = {datetime.fromisoformat(value) for value in values}
            if feature_dtype == "date":
                values = {date.fromisoformat(value) for value in values}

            non_categoricals[(feature_name, feature_dtype)] = values

    return non_categoricals


def _get_schemas_postgres(
    self: Artifact | Collection,
    related_data: dict | None = None,
) -> dict:
    if not related_data:
        artifact_meta = get_artifact_with_related(self, include_schema=True)
        related_data = artifact_meta.get("related_data", {})

    fs_data = related_data.get("schemas", {}) if related_data else {}
    return fs_data


def _create_feature_table(
    name: str, registry_str: str, data: list, show_header: bool = False
) -> Table:
    """Create a Rich table for a feature group."""
    table = Table(
        Column(name, style="", no_wrap=True, width=NAME_WIDTH),
        Column(registry_str, style="dim", no_wrap=True, width=TYPE_WIDTH),
        Column("", width=VALUES_WIDTH, no_wrap=True),
        show_header=show_header,
        box=None,
        pad_edge=False,
    )
    for row in data:
        table.add_row(*row)
    return table


def describe_features(
    self: Artifact,
    related_data: dict | None = None,
    to_dict: bool = False,
    tree: Tree | None = None,
    with_labels: bool = False,
):
    """Describe features of an artifact or collection."""
    from .artifact import Artifact

    # initialize tree
    if tree is None:
        tree = describe_header(self)

    dictionary: dict[str, Any] = {}

    if self._state.adding:
        return dictionary if to_dict else tree

    # feature sets
    schema_data: dict[str, tuple[str, list[str]]] = {}
    feature_data: dict[str, tuple[str, list[str]]] = {}
    if not to_dict:
        if self.id is not None and connections[self._state.db].vendor == "postgresql":
            fs_data = _get_schemas_postgres(self, related_data=related_data)
            for fs_id, (slot, data) in fs_data.items():
                for registry_str, feature_names in data.items():
                    # prevent projects show up as features
                    if registry_str == "Project":
                        continue
                    schema = Schema.objects.using(self._state.db).get(id=fs_id)
                    schema_data[slot] = (schema, feature_names)
                    for feature_name in feature_names:
                        feature_data[feature_name] = (slot, registry_str)
            schema_data.update(
                {
                    slot: (schema, schema.n)  # type: ignore
                    for slot, schema in get_schema_by_slot_(self).items()
                    if slot not in schema_data
                }
            )
        else:
            for slot, schema in get_schema_by_slot_(self).items():
                features = schema.members
                if features.exists():
                    # features.first() is a lot slower than features[0] here
                    name_field = get_name_field(features[0])
                    feature_names = list(
                        features.values_list(name_field, flat=True)[:20]
                    )
                    schema_data[slot] = (schema, feature_names)
                    for feature_name in feature_names:
                        feature_data[feature_name] = (slot, schema.itype)
                else:
                    schema_data[slot] = (schema, schema.n)

    internal_feature_names: dict[str, str] = {}
    if isinstance(self, Artifact):
        feature_sets = self.feature_sets.filter(itype="Feature")
        internal_feature_names = {}
        if len(feature_sets) > 0:
            for schema in feature_sets:
                internal_feature_names.update(
                    dict(schema.members.values_list("name", "dtype"))
                )

    # categorical feature values
    # Get the categorical data using the appropriate method
    if not self._state.adding and connections[self._state.db].vendor == "postgresql":
        categoricals = _get_categoricals_postgres(
            self,
            related_data=related_data,
        )
    else:
        categoricals = _get_categoricals(
            self,
        )

    # Get non-categorical features
    non_categoricals = _get_non_categoricals(
        self,
    )

    # Process all Features containing labels and sort into internal/external
    internal_feature_labels = {}
    external_data = []
    for features, is_list_type in [(categoricals, False), (non_categoricals, True)]:
        for (feature_name, feature_dtype), values in sorted(features.items()):
            # Handle dictionary conversion
            if to_dict:
                dict_value = values if len(values) > 1 else next(iter(values))
                dictionary[feature_name] = dict_value
                continue

            # Format message
            printed_values = (
                _format_values(sorted(values), n=10, quotes=False)
                if not is_list_type or not feature_dtype.startswith("list")
                else sorted(values)
            )

            # Sort into internal/external
            feature_info = (
                feature_name,
                Text(feature_dtype, style="dim"),
                printed_values,
            )
            if feature_name in internal_feature_names:
                internal_feature_labels[feature_name] = feature_info
            else:
                external_data.append(feature_info)

    if to_dict:
        return dictionary

    # Dataset features section
    # internal features that contain labels (only `Feature` features contain labels)
    internal_feature_labels_slot: dict[str, list] = {}
    for feature_name, feature_row in internal_feature_labels.items():
        slot, _ = feature_data.get(feature_name)
        internal_feature_labels_slot.setdefault(slot, []).append(feature_row)

    int_features_tree_children = []
    for slot, (schema, feature_names_or_n) in schema_data.items():
        if isinstance(feature_names_or_n, int):
            feature_rows = []
        else:
            feature_names = feature_names_or_n
            if slot in internal_feature_labels_slot:
                # add internal Feature features with labels
                feature_rows = internal_feature_labels_slot[slot]
                # add internal Feature features without labels
                feature_rows += [
                    (
                        feature_name,
                        Text(
                            str(internal_feature_names.get(feature_name)), style="dim"
                        ),
                        "",
                    )
                    for feature_name in feature_names
                    if feature_name and feature_name not in internal_feature_labels
                ]
            else:
                # add internal non-Feature features without labels
                feature_rows = [
                    (
                        feature_name,
                        Text(
                            str(
                                internal_feature_names.get(feature_name)
                                if feature_name in internal_feature_names
                                else schema.dtype
                            ),
                            style="dim",
                        ),
                        "",
                    )
                    for feature_name in feature_names
                    if feature_name
                ]
        int_features_tree_children.append(
            _create_feature_table(
                Text.assemble(
                    (slot, "violet"),
                    (" • ", "dim"),
                    (str(schema.n), "pink1"),
                ),
                Text.assemble((f"[{schema.itype}]", "pink1")),
                feature_rows,
                show_header=True,
            )
        )
    ## internal features from the non-`Feature` registry
    if int_features_tree_children:
        dataset_tree = tree.add(
            Text.assemble(
                ("Dataset features", "bold bright_magenta"),
            )
        )
        for child in int_features_tree_children:
            dataset_tree.add(child)

    # External features
    ext_features_tree_children = []
    if external_data:
        ext_features_tree_children.append(
            _create_feature_table(
                "",
                "",
                external_data,
            )
        )
    # ext_features_tree = None
    ext_features_header = Text("External features", style="bold dark_orange")
    if ext_features_tree_children:
        ext_features_tree = tree.add(ext_features_header)
        for child in ext_features_tree_children:
            ext_features_tree.add(child)
    if with_labels:
        # avoid querying the db if the labels were queried already
        labels_data = related_data.get("m2m") if related_data is not None else None
        labels_tree = describe_labels(self, labels_data=labels_data, as_subtree=True)
        if labels_tree:
            tree.add(labels_tree)

    return tree


def infer_feature_type_convert_json(
    key: str, value: Any, mute: bool = False
) -> tuple[str, Any, str]:
    from lamindb.base.dtypes import is_valid_datetime_str

    message = ""
    if isinstance(value, bool):
        return "bool", value, message
    elif isinstance(value, int):
        return "int", value, message
    elif isinstance(value, float):
        return "float", value, message
    elif isinstance(value, date):
        return "date", value.isoformat(), message
    elif isinstance(value, datetime):
        return "datetime", value.isoformat(), message
    elif isinstance(value, str):
        if datetime_str := is_valid_datetime_str(value):
            dt_type = (
                "date" if len(value) == 10 else "datetime"
            )  # YYYY-MM-DD is exactly 10 characters
            sanitized_value = datetime_str[:10] if dt_type == "date" else datetime_str  # type: ignore
            return dt_type, sanitized_value, message  # type: ignore
        else:
            return "cat ? str", value, message
    elif isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
        if isinstance(value, (pd.Series, np.ndarray, pd.Categorical)):
            dtype = serialize_pandas_dtype(value.dtype)
            if dtype == "str":
                # ndarray doesn't know categorical, so there was no conscious choice
                # offer both options
                if isinstance(value, np.ndarray):
                    dtype = "cat ? str"
                else:
                    # suggest to create a categorical if there are few unique values
                    message = suggest_categorical_for_str_iterable(value, key)
                    if message:
                        message = f"  # {message}"
            return dtype, list(value), message
        if isinstance(value, dict):
            return "dict", value, message
        if len(value) > 0:  # type: ignore
            first_element_type = type(next(iter(value)))
            if all(isinstance(elem, first_element_type) for elem in value):
                if first_element_type is bool:
                    return "list[bool]", value, message
                elif first_element_type is int:
                    return "list[int]", value, message
                elif first_element_type is float:
                    return "list[float]", value, message
                elif first_element_type is str:
                    return ("list[cat ? str]", value, message)
                elif first_element_type == SQLRecord:
                    return (
                        f"list[cat[{first_element_type.__get_name_with_module__()}]]",
                        value,
                        message,
                    )
    elif isinstance(value, SQLRecord):
        return (f"cat[{value.__class__.__get_name_with_module__()}]", value, message)
    if not mute:
        logger.warning(f"cannot infer feature type of: {value}, returning '?")
    return "?", value, message


def filter_base(
    queryset: BasicQuerySet,
    _skip_validation: bool = True,
    **expression,
) -> BasicQuerySet:
    from lamindb.models import Artifact, BasicQuerySet, QuerySet

    # not QuerySet but only BasicQuerySet
    assert isinstance(queryset, BasicQuerySet) and not isinstance(queryset, QuerySet)  # noqa: S101

    registry = queryset.model
    db = queryset.db

    model = Feature
    value_model = FeatureValue
    keys_normalized = [key.split("__")[0] for key in expression]
    if not _skip_validation:
        validated = model.using(db).validate(keys_normalized, field="name", mute=True)
        if sum(validated) != len(keys_normalized):
            raise ValidationError(
                f"Some keys in the filter expression are not registered as features: {np.array(keys_normalized)[~validated]}"
            )
    new_expression = {}
    features = model.using(db).filter(name__in=keys_normalized).distinct()
    feature_param = "feature"
    for key, value in expression.items():
        split_key = key.split("__")
        normalized_key = split_key[0]
        comparator = ""
        if len(split_key) == 2:
            comparator = f"__{split_key[1]}"
        feature = features.get(name=normalized_key)
        # non-categorical features
        if not feature.dtype.startswith("cat") and not feature.dtype.startswith(
            "list[cat"
        ):
            if comparator == "__isnull":
                if registry is Artifact:
                    from .artifact import ArtifactFeatureValue

                    if value:  # True
                        return queryset.exclude(
                            id__in=Subquery(
                                ArtifactFeatureValue.objects.filter(
                                    featurevalue__feature=feature
                                ).values("artifact_id")
                            )
                        )
                    else:
                        return queryset.exclude(
                            id__in=Subquery(
                                ArtifactFeatureValue.objects.filter(
                                    featurevalue__feature=feature
                                ).values("artifact_id")
                            )
                        )
            if comparator in {"__startswith", "__contains"}:
                logger.important(
                    f"currently not supporting `{comparator}`, using `__icontains` instead"
                )
                comparator = "__icontains"
            expression = {feature_param: feature, f"value{comparator}": value}
            feature_values = value_model.filter(**expression)
            new_expression[f"_{feature_param}_values__id__in"] = feature_values
        # categorical features
        elif isinstance(value, (str, SQLRecord, bool)):
            if comparator == "__isnull":
                if registry is Artifact:
                    result = parse_dtype(feature.dtype)[0]
                    kwargs = {
                        f"links_{result['registry'].__name__.lower()}__feature": feature
                    }
                    if value:  # True
                        return queryset.exclude(**kwargs)
                    else:
                        return queryset.filter(**kwargs)
            else:
                # because SQL is sensitive to whether querying with __in or not
                # and might return multiple equivalent records for the latter
                # we distinguish cases in which we have multiple label matches vs. one
                label = None
                labels = None
                result = parse_dtype(feature.dtype)[0]
                label_registry = result["registry"]
                if isinstance(value, str):
                    field_name = result["field"].field.name
                    # we need the comparator here because users might query like so
                    # ln.Artifact.filter(experiment__contains="Experi")
                    expression = {f"{field_name}{comparator}": value}
                    labels = result["registry"].using(db).filter(**expression)
                    if len(labels) == 0:
                        raise DoesNotExist(
                            f"Did not find a {label_registry.__name__} matching `{field_name}{comparator}={value}`"
                        )
                    elif len(labels) == 1:
                        label = labels[0]
                elif isinstance(value, SQLRecord):
                    label = value
                accessor_name = (
                    label_registry.artifacts.through.artifact.field._related_name
                )
                new_expression[f"{accessor_name}__feature"] = feature
                if label is not None:
                    # simplified query if we have exactly one label
                    new_expression[
                        f"{accessor_name}__{label_registry.__name__.lower()}"
                    ] = label
                else:
                    new_expression[
                        f"{accessor_name}__{label_registry.__name__.lower()}__in"
                    ] = labels
            # if passing a list of records, we want to
            # find artifacts that are annotated by all of them at the same
            # time; hence, we don't want the __in construct that we use to match strings
            # https://laminlabs.slack.com/archives/C04FPE8V01W/p1688328084810609
    if not new_expression:
        raise NotImplementedError
    return queryset.filter(**new_expression)


def filter_with_features(
    queryset: BasicQuerySet, *queries, **expressions
) -> BasicQuerySet:
    from lamindb.models import Artifact, BasicQuerySet, QuerySet

    if isinstance(queryset, QuerySet):
        # need to avoid infinite recursion because
        # filter_with_features is called in queryset.filter otherwise
        filter_kwargs = {"_skip_filter_with_features": True}
    else:
        filter_kwargs = {}

    registry = queryset.model

    if registry is Artifact and not any(e.startswith("kind") for e in expressions):
        exclude_kwargs = {"kind": "__lamindb_run__"}
    else:
        exclude_kwargs = {}

    if expressions:
        keys_normalized = [key.split("__")[0] for key in expressions]
        field_or_feature = keys_normalized[0]
        if field_or_feature in registry.__get_available_fields__():
            qs = queryset.filter(*queries, **expressions, **filter_kwargs)
        elif all(
            features_validated := Feature.objects.using(queryset.db).validate(
                keys_normalized, field="name", mute=True
            )
        ):
            # filter_base requires qs to be BasicQuerySet
            qs = filter_base(
                queryset._to_class(BasicQuerySet, copy=True),
                _skip_validation=True,
                **expressions,
            )._to_class(type(queryset), copy=False)
            qs = qs.filter(*queries, **filter_kwargs)
        else:
            features = ", ".join(sorted(np.array(keys_normalized)[~features_validated]))
            message = f"feature names: {features}"
            avail_fields = registry.__get_available_fields__()
            if "_branch_code" in avail_fields:
                avail_fields.remove("_branch_code")  # backward compat
            fields = ", ".join(sorted(avail_fields))
            raise InvalidArgument(
                f"You can query either by available fields: {fields}\n"
                f"Or fix invalid {message}"
            )
    else:
        qs = queryset.filter(*queries, **filter_kwargs)

    return qs.exclude(**exclude_kwargs) if exclude_kwargs else qs


# for deprecated functionality
def _unify_staged_feature_sets_by_hash(
    feature_sets: MutableMapping[str, Schema],
):
    unique_values: dict[str, Any] = {}

    for key, value in feature_sets.items():
        value_hash = value.hash  # Assuming each value has a .hash attribute
        if value_hash in unique_values:
            feature_sets[key] = unique_values[value_hash]
        else:
            unique_values[value_hash] = value

    return feature_sets


# for deprecated functionality
def parse_staged_feature_sets_from_anndata(
    adata: AnnData,
    var_field: FieldAttr | None = None,
    obs_field: FieldAttr = Feature.name,
    uns_field: FieldAttr | None = None,
    mute: bool = False,
    organism: str | SQLRecord | None = None,
) -> dict:
    data_parse = adata
    if not isinstance(adata, AnnData):  # is a path
        filepath = create_path(adata)  # returns Path for local
        if not isinstance(filepath, LocalPathClasses):
            from lamindb import settings
            from lamindb.core.storage._backed_access import backed_access

            using_key = settings._using_key
            data_parse = backed_access(filepath, using_key=using_key)
        else:
            data_parse = ad.read_h5ad(filepath, backed="r")
        dtype = "float"
    else:
        dtype = "float" if adata.X is None else serialize_pandas_dtype(adata.X.dtype)
    feature_sets = {}
    if var_field is not None:
        schema_var = Schema.from_values(
            data_parse.var.index,
            var_field,
            dtype=dtype,
            mute=mute,
            organism=organism,
            raise_validation_error=False,
        )
        if schema_var is not None:
            feature_sets["var"] = schema_var
    if obs_field is not None and len(data_parse.obs.columns) > 0:
        schema_obs = Schema.from_dataframe(
            df=data_parse.obs,
            field=obs_field,
            mute=mute,
            organism=organism,
        )
        if schema_obs is not None:
            feature_sets["obs"] = schema_obs
    if uns_field is not None and len(data_parse.uns) > 0:
        validated_features = Feature.from_values(  # type: ignore
            data_parse.uns.keys(), field=uns_field, organism=organism
        )
        if len(validated_features) > 0:
            schema_uns = Schema(validated_features, dtype=None, otype="dict")
            feature_sets["uns"] = schema_uns
    return feature_sets


class FeatureManager:
    """Feature manager."""

    def __init__(self, host: Artifact | Run):
        self._host = host
        self._slots: dict[str, Schema] | None = None
        self._accessor_by_registry_ = None

    def __repr__(self) -> str:
        return self.describe(return_str=True)  # type: ignore

    def describe(self, return_str: bool = False) -> str | None:
        tree = describe_features(self._host)  # type: ignore
        return format_rich_tree(
            tree, fallback="no linked features", return_str=return_str
        )

    def get_values(self) -> dict[str, Any]:
        """Get feature values as a dictionary."""
        return describe_features(self._host, to_dict=True)  # type: ignore

    @deprecated("slots[slot].members")
    def __getitem__(self, slot) -> BasicQuerySet:
        if slot not in self.slots:
            raise ValueError(
                f"No linked feature set for slot: {slot}\nDid you get validation"
                " warnings? Only features that match registered features get validated"
                " and linked."
            )
        schema = self.slots[slot]
        orm_name = schema.itype
        return getattr(schema, self._accessor_by_registry[orm_name]).all()

    @property
    def slots(self) -> dict[str, Schema]:
        """Schema by slot.

        Example::

            artifact.features.slots
            #> {'var': <Schema: var>, 'obs': <Schema: obs>}
        """
        if self._slots is None:
            self._slots = get_schema_by_slot_(self._host)
        return self._slots

    @property
    def _accessor_by_registry(self):
        """Accessor by registry."""
        if self._accessor_by_registry_ is None:
            self._accessor_by_registry_ = get_accessor_by_registry_(self._host)
        return self._accessor_by_registry_

    def _add_label_feature_links(
        self,
        features_labels,
        *,
        label_ref_is_name: bool | None = None,
        feature_ref_is_name: bool | None = None,
    ):
        if list(features_labels.keys()) != ["ULabel"]:
            related_names = dict_related_model_to_related_name(self._host.__class__)
        else:
            related_names = {"ULabel": "ulabels"}
        for class_name, registry_features_labels in features_labels.items():
            related_name = related_names[class_name]  # e.g., "ulabels"
            IsLink = getattr(self._host, related_name).through
            field_name = f"{get_link_attr(IsLink, self._host)}_id"  # e.g., ulabel_id
            links = [
                IsLink(
                    **{
                        "artifact_id": self._host.id,
                        "feature_id": feature.id,
                        field_name: label.id,
                        "feature_ref_is_name": feature_ref_is_name,
                        "label_ref_is_name": label_ref_is_name,
                    }
                )
                for (feature, label) in registry_features_labels
            ]
            # a link might already exist
            try:
                save(links, ignore_conflicts=False)
            except Exception:
                save(links, ignore_conflicts=True)
            # now delete links that were previously saved without a feature
            IsLink.filter(
                **{
                    "artifact_id": self._host.id,
                    "feature_id": None,
                    f"{field_name}__in": [l.id for _, l in registry_features_labels],
                }
            ).delete()

    def add_values(
        self,
        values: dict[str, str | int | float | bool],
        feature_field: FieldAttr = Feature.name,
        schema: Schema = None,
    ) -> None:
        """Curate artifact with features & values.

        Args:
            values: A dictionary of keys (features) & values (labels, numbers, booleans).
            feature_field: The field of a reference registry to map keys of the dictionary.
            schema: Schema to validate against.
        """
        from lamindb.base.dtypes import is_iterable_of_sqlrecord

        from .._tracked import get_current_tracked_run

        # rename to distinguish from the values inside the dict
        dictionary = values
        keys = dictionary.keys()
        if isinstance(keys, DICT_KEYS_TYPE):
            keys = list(keys)  # type: ignore
        # deal with other cases later
        assert all(isinstance(key, str) for key in keys)  # noqa: S101

        registry = feature_field.field.model
        value_model = FeatureValue
        model_name = "Feature"

        if schema is not None:
            from lamindb.curators.core import ExperimentalDictCurator

            ExperimentalDictCurator(values, schema).validate()
            records = schema.members.filter(name__in=keys)
        else:
            records = registry.from_values(keys, field=feature_field, mute=True)
            if len(records) != len(keys):
                not_validated_keys = [
                    key for key in keys if key not in records.to_list("name")
                ]
                not_validated_keys_dtype_message = [
                    (key, infer_feature_type_convert_json(key, dictionary[key]))
                    for key in not_validated_keys
                ]
                run = get_current_tracked_run()
                if run is not None:
                    name = f"{run.transform.type}[{run.transform.key}]"
                    type_hint = f"""  {model_name.lower()}_type = ln.{model_name}(name='{name}', is_type=True).save()"""
                    elements = [type_hint]
                    type_kwarg = f", type={model_name.lower()}_type"
                else:
                    elements = []
                    type_kwarg = ""
                elements += [
                    f"  ln.{model_name}(name='{key}', dtype='{dtype}'{type_kwarg}).save(){message}"
                    for key, (dtype, _, message) in not_validated_keys_dtype_message
                ]
                hint = "\n".join(elements)
                msg = (
                    f"These keys could not be validated: {not_validated_keys}\n"
                    f"Here is how to create a {model_name.lower()}:\n\n{hint}"
                )
                raise ValidationError(msg)

        features_labels = defaultdict(list)
        _feature_values = []
        not_validated_values: dict[str, list[str]] = defaultdict(list)
        for feature in records:
            value = dictionary[feature.name]
            inferred_type, converted_value, _ = infer_feature_type_convert_json(
                feature.name,
                value,
                mute=True,
            )
            if feature.dtype == "num":
                if inferred_type not in {"int", "float"}:
                    raise TypeError(
                        f"Value for feature '{feature.name}' with dtype {feature.dtype} must be a number, but is {value} with dtype {inferred_type}"
                    )
            elif feature.dtype.startswith("cat"):
                if inferred_type != "?":
                    if not (
                        inferred_type.startswith("cat")
                        or inferred_type == "list[cat ? str]"
                        or isinstance(value, SQLRecord)
                        or is_iterable_of_sqlrecord(value)
                    ):
                        raise TypeError(
                            f"Value for feature '{feature.name}' with dtype '{feature.dtype}' must be a string or record, but is {value} with dtype {inferred_type}"
                        )
            elif (
                (feature.dtype == "str" and inferred_type != "cat ? str")
                or (feature.dtype == "list[str]" and inferred_type != "list[cat ? str]")
                or (
                    feature.dtype.startswith("list[cat")
                    and inferred_type != "list[cat ? str]"
                )
                or (
                    feature.dtype not in {"str", "list[str]"}
                    and not feature.dtype.startswith("list[cat")
                    and feature.dtype != inferred_type
                )
            ):
                raise ValidationError(
                    f"Expected dtype for '{feature.name}' is '{feature.dtype}', got '{inferred_type}'"
                )
            if not (
                feature.dtype.startswith("cat") or feature.dtype.startswith("list[cat")
            ):
                filter_kwargs = {model_name.lower(): feature, "value": converted_value}
                feature_value, _ = value_model.get_or_create(**filter_kwargs)
                _feature_values.append(feature_value)
            else:
                if isinstance(value, SQLRecord) or is_iterable_of_sqlrecord(value):
                    if isinstance(value, SQLRecord):
                        label_records = [value]
                    else:
                        label_records = value  # type: ignore
                    for record in label_records:
                        if record._state.adding:
                            raise ValidationError(
                                f"Please save {record} before annotation."
                            )
                        features_labels[
                            record.__class__.__get_name_with_module__()
                        ].append((feature, record))
                else:
                    if isinstance(value, str):
                        values = [value]  # type: ignore
                    else:
                        values = value  # type: ignore
                    if feature.dtype == "cat":
                        feature.dtype += "[ULabel]"
                        feature.save()
                        result = {
                            "registry_str": "ULabel",
                            "registry": ULabel,
                            "field": ULabel.name,
                        }
                    else:
                        result = parse_dtype(feature.dtype)[0]
                    validated = result["registry"].validate(  # type: ignore
                        values, field=result["field"], mute=True
                    )
                    values_array = np.array(values)
                    validated_values = values_array[validated]
                    if validated.sum() != len(values):
                        not_validated_values[result["registry_str"]] += values_array[  # type: ignore
                            ~validated
                        ].tolist()
                    label_records = result["registry"].from_values(  # type: ignore
                        validated_values, field=result["field"], mute=True
                    )
                    features_labels[result["registry_str"]] += [  # type: ignore
                        (feature, label_record) for label_record in label_records
                    ]
        if not_validated_values:
            hint = ""
            for key, values_list in not_validated_values.items():
                key_str = "ln.ULabel" if key == "ULabel" else key
                hint += f"  records = {key_str}.from_values({values_list}, create=True).save()\n"
            msg = (
                f"These values could not be validated: {dict(not_validated_values)}\n"
                f"Here is how to create records for them:\n\n{hint}"
            )
            raise ValidationError(msg)

        if features_labels:
            self._add_label_feature_links(features_labels)
        if _feature_values:
            to_insert_feature_values = [
                record for record in _feature_values if record._state.adding
            ]
            if to_insert_feature_values:
                save(to_insert_feature_values)
            dict_typed_features = [
                getattr(record, model_name.lower())
                for record in _feature_values
                if getattr(record, model_name.lower()).dtype == "dict"
            ]
            IsLink = self._host._feature_values.through
            valuefield_id = "featurevalue_id"
            host_class_lower = self._host.__class__.__get_name_with_module__().lower()
            if dict_typed_features:
                # delete all previously existing anotations with dictionaries
                kwargs = {
                    f"links_{host_class_lower}__{host_class_lower}_id": self._host.id,
                    f"{model_name.lower()}__in": dict_typed_features,
                }
                try:
                    value_model.filter(**kwargs).delete()
                except ProtectedError:
                    pass
            # add new feature links
            links = [
                IsLink(
                    **{
                        f"{host_class_lower}_id": self._host.id,
                        valuefield_id: feature_value.id,
                    }
                )
                for feature_value in _feature_values
            ]
            # a link might already exist, to avoid raising a unique constraint
            # error, ignore_conflicts
            save(links, ignore_conflicts=True)

    def remove_values(
        self,
        feature: str | Feature,
        *,
        value: Any | None = None,
    ) -> None:
        """Remove value annotations for a given feature.

        Args:
            feature: The feature for which to remove values.
            value: An optional value to restrict removal to a single value.

        """
        from .artifact import Artifact

        if isinstance(feature, str):
            feature = Feature.get(name=feature)
        filter_kwargs = {"feature": feature}
        if feature.dtype.startswith("cat["):  # type: ignore
            feature_registry = feature.dtype.replace("cat[", "").replace("]", "")  # type: ignore
            if value is not None:
                assert isinstance(value, SQLRecord)  # noqa: S101
                # the below uses our convention for field names in link models
                link_name = (
                    feature_registry.split(".")[1]
                    if "." in feature_registry
                    else feature_registry
                ).lower()
                filter_kwargs[link_name] = value
            if feature_registry == "ULabel":
                link_attribute = "links_ulabel"
            else:
                link_models_on_models = {
                    getattr(
                        Artifact, obj.related_name
                    ).through.__get_name_with_module__(): obj.related_model.__get_name_with_module__()
                    for obj in Artifact._meta.related_objects
                    if obj.related_model.__get_name_with_module__() == feature_registry
                }
                link_attribute = {
                    obj.related_name
                    for obj in Artifact._meta.related_objects
                    if obj.related_model.__get_name_with_module__()
                    in link_models_on_models
                }.pop()
            getattr(self._host, link_attribute).filter(**filter_kwargs).delete()
        else:
            if value is not None:
                filter_kwargs["value"] = value
            feature_values = self._host._feature_values.filter(**filter_kwargs)
            self._host._feature_values.remove(*feature_values)
            # this might leave a dangling feature_value record
            # but we don't want to pay the price of making another query just to remove this annotation
            # we can clean the FeatureValue registry periodically if we want to

    def _add_schema(self, schema: Schema, slot: str) -> None:
        """Annotate artifact with a schema.

        Args:
            schema: `Schema` A schema record.
            slot: `str` The slot that marks where the schema is stored in
                the artifact.
        """
        # TODO: deprecate as soon as we have the Schema-based curators
        if self._host._state.adding:
            raise ValueError(
                "Please save the artifact or collection before adding a feature set!"
            )
        host_db = self._host._state.db
        schema.save(using=host_db)
        kwargs = {
            "artifact_id": self._host.id,
            "schema": schema,
            "slot": slot,
        }
        link_record = (
            self._host.feature_sets.through.objects.using(host_db)
            .filter(**kwargs)
            .one_or_none()
        )
        if link_record is None:
            self._host.feature_sets.through(**kwargs).save(using=host_db)
            if slot in self.slots:
                logger.debug(f"replaced existing {slot} feature set")
            self._slots[slot] = schema  # type: ignore

    def _add_from(self, data: Artifact | Collection, transfer_logs: dict = None):
        """Transfer features from a artifact or collection."""
        # This only covers feature sets
        if transfer_logs is None:
            transfer_logs = {"mapped": [], "transferred": [], "run": None}
        from lamindb import settings

        using_key = settings._using_key
        for slot, schema in data.features.slots.items():  # type: ignore
            try:
                members = schema.members
            except ModuleWasntConfigured as err:
                logger.warning(f"skipping transfer of {slot} schema because {err}")
                continue
            if len(members) == 0:
                continue
            if len(members) > settings.annotation.n_max_records:
                logger.warning(
                    f"skipping creating {len(members)} > {settings.annotation.n_max_records} new {members[0].__class__.__name__} records"
                )
                schema_self = schema
                schema_exists = Schema.filter(hash=schema_self.hash).one_or_none()
                if schema_exists is not None:
                    schema_self = schema_exists
                else:
                    schema_self.save()
            else:
                registry = members[0].__class__
                # note here the features are transferred based on an unique field
                field = REGISTRY_UNIQUE_FIELD.get(registry.__name__.lower(), "uid")
                # this will be e.g. be a list of ontology_ids or uids
                member_uids = list(members.values_list(field, flat=True))
                validated = registry.validate(member_uids, field=field, mute=True)
                new_members_uids = list(compress(member_uids, ~validated))
                new_members = members.filter(**{f"{field}__in": new_members_uids})
                n_new_members = len(new_members)
                if len(members) > settings.annotation.n_max_records:
                    logger.warning(
                        f"skipping creating {n_new_members} > {settings.annotation.n_max_records} new {registry.__name__} records"
                    )
                if n_new_members > 0:
                    # transfer foreign keys needs to be run before transfer to default db
                    transfer_fk_to_default_db_bulk(
                        new_members, using_key, transfer_logs=transfer_logs
                    )
                    for feature in new_members:
                        # not calling save=True here as in labels, because want to
                        # bulk save below
                        # transfer_fk is set to False because they are already transferred
                        # in the previous step transfer_fk_to_default_db_bulk
                        transfer_to_default_db(
                            feature,
                            using_key,
                            transfer_fk=False,
                            transfer_logs=transfer_logs,
                        )
                    save(
                        new_members, ignore_conflicts=True
                    )  # conflicts arising from existing records are ignored

                # create a new feature set from feature values using the same uid
                schema_self = Schema.from_values(
                    member_uids, field=getattr(registry, field)
                )
                if schema_self is None:
                    if hasattr(registry, "organism_id"):
                        logger.warning(
                            f"Schema is not transferred, check if organism is set correctly: {schema}"
                        )
                    continue
                # make sure the uid matches if schema is composed of same features
                if schema_self.hash == schema.hash:
                    schema_self.uid = schema.uid
                logger.info(f"saving {slot} schema: {schema_self}")
            try:
                self._host.features._add_schema(schema_self, slot)
            except IntegrityError:
                logger.warning(
                    f"updating annotation of artifact {self._host.uid} with feature set for slot: {slot}"
                )
                self._host.feature_sets.through.objects.get(
                    artifact_id=self._host.id, slot=slot
                ).delete()
                self._host.features._add_schema(schema_self, slot)

    def make_external(self, feature: Feature) -> None:
        """Make a feature external, aka, remove feature from feature sets.

        Args:
            feature: `Feature` A feature record.

        """
        if not isinstance(feature, Feature):
            raise TypeError("feature must be a Feature record!")
        feature_sets = Schema.filter(features=feature)
        for fs in feature_sets:
            f = Feature.filter(uid=feature.uid)
            features_updated = fs.members.difference(f)
            if len(features_updated) > 0:
                # re-compute the hash of feature sets based on the updated members
                features_hash = hash_set({feature.uid for feature in features_updated})
                fs.hash = features_hash
                fs.n = len(features_updated)
                fs.save()
            # delete the link between the feature and the feature set
            Schema.features.through.objects.filter(
                feature_id=feature.id, schema_id=fs.id
            ).delete()
            # if no members are left in the schema, delete it
            if len(features_updated) == 0:
                logger.warning(f"deleting empty feature set: {fs}")
                fs.artifacts.set([])
                fs.delete()

    @deprecated("_add_schema")
    def add_schema(self, schema: Schema, slot: str) -> None:
        return self._add_schema(schema, slot)

    @deprecated("_add_schema")
    def add_feature_set(self, schema: Schema, slot: str) -> None:
        return self._add_schema(schema, slot)

    @property
    @deprecated("slots")
    def _schema_by_slot(self):
        return self.slots

    @property
    def _feature_set_by_slot(self):
        return self.slots

    # no longer called from within curator
    # deprecated
    def _add_set_from_df(
        self,
        field: FieldAttr = Feature.name,
        organism: str | None = None,
        mute: bool = False,
    ):
        """Add feature set corresponding to column names of DataFrame."""
        assert self._host.otype == "DataFrame"  # noqa: S101
        df = self._host.load(is_run_input=False)
        schema = Schema.from_dataframe(
            df=df,
            field=field,
            mute=mute,
            organism=organism,
        )
        self._host._staged_feature_sets = {"columns": schema}
        self._host.save()

    # deprecated
    def _add_set_from_anndata(
        self,
        var_field: FieldAttr | None = None,
        obs_field: FieldAttr | None = Feature.name,
        uns_field: FieldAttr | None = None,
        mute: bool = False,
        organism: str | SQLRecord | None = None,
    ):
        """Add features from AnnData."""
        assert self._host.otype == "AnnData"  # noqa: S101

        # parse and register features
        adata = self._host.load(is_run_input=False)
        feature_sets = parse_staged_feature_sets_from_anndata(
            adata,
            var_field=var_field,
            obs_field=obs_field,
            uns_field=uns_field,
            mute=mute,
            organism=organism,
        )

        # link feature sets
        self._host._staged_feature_sets = feature_sets
        self._host.save()

    # deprecated
    def _add_set_from_mudata(
        self,
        var_fields: dict[str, FieldAttr] | None = None,
        obs_fields: dict[str, FieldAttr] | None = None,
        mute: bool = False,
        organism: str | SQLRecord | None = None,
    ):
        """Add features from MuData."""
        if obs_fields is None:
            obs_fields = {}
        assert self._host.otype == "MuData"  # noqa: S101

        # parse and register features
        mdata = self._host.load(is_run_input=False)
        feature_sets = {}

        obs_features = Feature.from_values(mdata.obs.columns)  # type: ignore
        if len(obs_features) > 0:
            feature_sets["obs"] = Schema(features=obs_features)
        for modality, field in var_fields.items():
            modality_fs = parse_staged_feature_sets_from_anndata(
                mdata[modality],
                var_field=field,
                obs_field=obs_fields.get(modality, Feature.name),
                mute=mute,
                organism=organism,
            )
            for k, v in modality_fs.items():
                feature_sets[f"['{modality}'].{k}"] = v

        # link feature sets
        self._host._staged_feature_sets = _unify_staged_feature_sets_by_hash(
            feature_sets
        )
        self._host.save()

    # deprecated
    def _add_set_from_spatialdata(
        self,
        sample_metadata_key: str,
        sample_metadata_field: FieldAttr = Feature.name,
        var_fields: dict[str, FieldAttr] | None = None,
        obs_fields: dict[str, FieldAttr] | None = None,
        mute: bool = False,
        organism: str | SQLRecord | None = None,
    ):
        """Add features from SpatialData."""
        obs_fields, var_fields = obs_fields or {}, var_fields or {}
        assert self._host.otype == "SpatialData"  # noqa: S101

        # parse and register features
        sdata = self._host.load(is_run_input=False)
        feature_sets = {}

        # sample features
        sample_features = Feature.from_values(
            sdata.get_attrs(
                key=sample_metadata_key, return_as="df", flatten=True
            ).columns,
            field=sample_metadata_field,
        )  # type: ignore
        if len(sample_features) > 0:
            feature_sets[sample_metadata_key] = Schema(features=sample_features)

        # table features
        for table, field in var_fields.items():
            table_fs = parse_staged_feature_sets_from_anndata(
                sdata[table],
                var_field=field,
                obs_field=obs_fields.get(table, Feature.name),
                mute=mute,
                organism=organism,
            )
            for k, v in table_fs.items():
                feature_sets[f"['{table}'].{k}"] = v

        # link feature sets
        self._host._staged_feature_sets = _unify_staged_feature_sets_by_hash(
            feature_sets
        )
        self._host.save()



# Generated by Django 5.2 on 2025-06-30 12:54

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0107_add_schema_to_record"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="record",
            name="sheet",
        ),
        migrations.RemoveField(
            model_name="project",
            name="sheets",
        ),
        migrations.AlterUniqueTogether(
            name="sheetproject",
            unique_together=None,
        ),
        migrations.DeleteModel(
            name="Sheet",
        ),
        migrations.DeleteModel(
            name="SheetProject",
        ),
    ]



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    overload,
)
from uuid import UUID

from django.db import models
from lamin_utils import logger
from lamindb_setup import settings as setup_settings
from lamindb_setup.core._hub_core import (
    delete_storage_record,
    get_storage_records_for_instance,
    select_space,
    update_storage_with_space,
)
from lamindb_setup.core._settings_storage import (
    StorageSettings,
    get_storage_type,
    init_storage,
)
from lamindb_setup.core.upath import check_storage_is_empty, create_path

from lamindb.base.fields import (
    CharField,
    TextField,
)

from ..base.ids import base62_12
from .run import TracksRun, TracksUpdates
from .sqlrecord import Space, SQLRecord

if TYPE_CHECKING:
    from pathlib import Path

    from lamindb_setup.types import StorageType
    from upath import UPath

    from .artifact import Artifact


class Storage(SQLRecord, TracksRun, TracksUpdates):
    """Storage locations of artifacts such as local directories or S3 buckets.

    A storage location is either a directory (local or a folder in the cloud) or
    an entire S3/GCP bucket.
    A LaminDB instance can manage and read from multiple storage locations. But any
    storage location is managed by *at most one* LaminDB instance.

    .. dropdown:: Managed vs. read-only storage locations

        A LaminDB instance can only write artifacts to its managed storage
        locations.

        The :attr:`~lamindb.Storage.instance_uid` field defines the managing LaminDB instance of a storage location.
        You can access the `instance_uid` of your current instance through `ln.setup.settings.instance_uid`.

        Here is an example (`source <https://lamin.ai/laminlabs/lamindata/transform/dPco79GYgzag0000>`__).

        .. image:: https://lamin-site-assets.s3.amazonaws.com/.lamindb/eHDmIOAxLEoqZ2oK0000.png
           :width: 400px

        Some public storage locations are not be managed by any LaminDB instance: their `instance_uid` is `None`.

    .. dropdown:: Managing access to storage locations across instances

        You can manage access through LaminHub's fine-grained access management or
        through AWS policies that you attach to your S3 bucket.

        To enable access management via LaminHub, head over to `https://lamin.ai/{account}/infrastructure`.
        By clicking the green button that says "Connect S3 bucket", LaminDB will start connecting through federated S3 tokens
        so that your collaborators access data based on their permissions in LaminHub.
        :doc:`docs:access` has more details.

        .. image:: https://lamin-site-assets.s3.amazonaws.com/.lamindb/ze8hkgVxVptSSZEU0000.png
           :width: 800px

        By default, access permissions to a storage location are governed by the access permissions of its managing instance. If you
        want to further restrict access to a storage location, you can move it into a space::

            space = ln.Space.get(name="my-space")
            storage_loc = ln.Storage.get(root="s3://my-storace-location")
            storage_loc.space = space
            storage_loc.save()

        If you don't want to store data in the cloud, you can use local storage locations: :doc:`faq/keep-artifacts-local`.

    Args:
        root: `str` The root path of the storage location, e.g., `"./mydir"`, `"s3://my-bucket"`, `"s3://my-bucket/myfolder"`, `"gs://my-bucket/myfolder"`, `"/nfs/shared/datasets/genomics"`, `"/weka/shared/models/"`, ...
        description: `str | None = None` An optional description.
        space: `Space | None = None` A space to restrict access permissions to the storage location.
        host: `str | None = None` For local storage locations, pass a globally unique host identifier, e.g. `"my-institute-cluster-1"`, `"my-server-abcd"`, ...

    See Also:
        :attr:`lamindb.core.Settings.storage`
            Current default storage location of your compute session for writing artifacts.
        :attr:`~lamindb.setup.core.StorageSettings`
            Storage settings.
        :doc:`faq/keep-artifacts-local`
            Avoid storing artifacts in the cloud, but keep them on local infrastructure.

    Examples:

        When you create a LaminDB instance, you configure its default storage location via `--storage`::

            lamin init --storage ./mydatadir  # or "s3://my-bucket/myfolder", "gs://my-bucket/myfolder", ...

        View the current default storage location for writing artifacts::

            import lamindb as ln

            ln.settings.storage

        Create a new cloud storage location::

            ln.Storage(root="s3://our-bucket/our-folder").save()

        Create a new local storage location::

            ln.Storage(root="/dir/our-shared-dir", host="our-server-123").save()

        Globally switch to another storage location::

            ln.settings.storage = "/dir/our-shared-dir"  # or "s3://our-bucket/our-folder", "gs://our-bucket/our-folder", ...

        Or if you're operating in `keep-artifacts-local` mode (:doc:`faq/keep-artifacts-local`)::

            ln.settings.local_storage = "/dir/our-other-shared-dir"

        View all storage locations used in your LaminDB instance::

            ln.Storage.to_dataframe()

    Notes:

        .. dropdown:: What is the `.lamindb/` directory inside a storage location?

            It stores all artifacts that are ingested through `lamindb`, indexed by the artifact `uid`.
            This means you don't have to worry about renaming or moving files, as this all happens on the database level.

            Existing artifacts are typically stored in hierarchical structures with semantic folder names.
            Instead of copying such artifacts into `.lamindb/` upon calls of `Artifact("legacy_path").save()`,
            LaminDB registers them with the semantic `key` representing the relative path within the storage location.
            These artifacts are marked with `artifact._key_is_virtual = False` and treated correspondingly.

            There is only a single `.lamindb/` directory per storage location.

        .. dropdown:: What should I do if I want to bulk migrate all artifacts to another storage?

            Currently, you can only achieve this manually and you should be careful with it.

            1. Copy or move artifacts into the desired new storage location
            2. Adapt the corresponding record in the {class}`~lamindb.Storage` registry by setting the `root` field to the new location
            3. If your LaminDB storage location is managed through the hub, you also need to update the storage record on the hub -- contact support

    """

    class Meta(SQLRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "lamindb"

    _name_field: str = "root"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False, unique=True, max_length=12, default=base62_12, db_index=True
    )
    """Universal id, valid across DB instances."""
    root: str = CharField(db_index=True, unique=True)
    """Root path of storage (cloud or local path)."""
    description: str | None = TextField(null=True)
    """A description."""
    type: StorageType = CharField(max_length=30, db_index=True)
    """Can be "local" vs. "s3" vs. "gs". Is auto-detected from the format of the `root` path."""
    region: str | None = CharField(max_length=64, db_index=True, null=True)
    """Storage region for cloud storage locations. Host identifier for local storage locations."""
    instance_uid: str | None = CharField(max_length=12, db_index=True, null=True)
    """Instance that manages this storage location."""
    artifacts: Artifact
    """Artifacts contained in this storage location."""

    @overload
    def __init__(
        self,
        root: str,
        *,
        description: str | None = None,
        space: Space | None = None,
        host: str | None = None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args)
            self._old_space = self.space
            self._old_space_id = self.space_id
            return None
        if args:
            assert len(args) == 1, (  # noqa: S101
                "Storage can only be initialized with a single positional argument, the root path."
            )
            kwargs["root"] = args[0]
        if "host" in kwargs:
            if "type" in kwargs:
                assert kwargs["type"] == "local", (  # noqa: S101
                    "type needs to be 'local' if host is set"
                )
            else:
                kwargs["type"] = "local"
            assert get_storage_type(kwargs["root"]) == "local", (  # noqa: S101
                "root must be a local path if host is set"
            )
            assert "region" not in kwargs, "region must not be set if host is set"  # noqa: S101
            kwargs["region"] = kwargs.pop("host")
            storage_record = Storage.filter(
                root=kwargs["root"], region=kwargs["region"]
            ).one_or_none()
        else:
            storage_record = Storage.filter(root=kwargs["root"]).one_or_none()
        space = kwargs.get("space", None)
        if storage_record is not None:
            from .sqlrecord import init_self_from_db

            init_self_from_db(self, storage_record)
            self._old_space = self.space
            self._old_space_id = self.space_id
            return None

        skip_preparation = kwargs.pop("_skip_preparation", False)
        if skip_preparation:
            assert space is None, "`space` must not be set if _skip_preparation is True"  # noqa: S101
            super().__init__(*args, **kwargs)
            return None

        space_uuid = None
        if space is not None:
            hub_space_record = select_space(space.uid)
            if hub_space_record is None:
                raise ValueError(
                    "Please first create a space on the hub: https://docs.lamin.ai/access"
                )
            space_uuid = UUID(hub_space_record["id"])

        # instance_id won't take effect if
        # - there is no write access
        # - the storage location is already managed by another instance
        ssettings, _ = init_storage(
            kwargs["root"],
            instance_id=setup_settings.instance._id,
            instance_slug=setup_settings.instance.slug,
            register_hub=setup_settings.instance.is_on_hub,
            region=kwargs.get("region", None),  # host was renamed to region already
            space_uuid=space_uuid,
        )
        # ssettings performed validation and normalization of the root path
        kwargs["root"] = ssettings.root_as_str  # noqa: S101
        if "instance_uid" in kwargs:
            assert kwargs["instance_uid"] == ssettings.instance_uid  # noqa: S101
        else:
            kwargs["instance_uid"] = ssettings.instance_uid
        if ssettings._uid is not None:  # need private attribute here
            kwargs["uid"] = ssettings._uid
        if "type" not in kwargs:
            kwargs["type"] = ssettings.type
        else:
            assert kwargs["type"] == ssettings.type  # noqa: S101
        if "region" in kwargs:
            assert kwargs["region"] == ssettings.region  # noqa: S101
        else:
            kwargs["region"] = ssettings.region

        is_managed_by_current_instance = (
            ssettings.instance_uid == setup_settings.instance.uid
        )
        if ssettings.instance_uid is not None and not is_managed_by_current_instance:
            is_managed_by_instance = (
                f", is managed by instance with uid {ssettings.instance_uid}"
            )
        else:
            is_managed_by_instance = ""
        hub_message = ""
        if setup_settings.instance.is_on_hub and is_managed_by_current_instance:
            instance_owner = setup_settings.instance.owner
            ui_url = setup_settings.instance.ui_url
            hub_message = f", see: {ui_url}/{instance_owner}/infrastructure"
        managed_message = (
            "created managed"
            if is_managed_by_current_instance
            else "referenced read-only"
        )
        logger.important(
            f"{managed_message} storage location at {kwargs['root']}{is_managed_by_instance}{hub_message}"
        )
        super().__init__(**kwargs)
        self._old_space = self.space
        self._old_space_id = self.space_id

    @property
    def host(self) -> str | None:
        """Host identifier for local storage locations.

        Is `None` for locations with `type != "local"`.

        A globally unique user-defined host identifier (cluster, server, laptop, etc.).
        """
        if self.type != "local":
            return None
        return self.region

    @property
    def path(self) -> Path | UPath:
        """Path.

        Uses the `.root` field and converts it into a `Path` or `UPath`.
        """
        access_token = self._access_token if hasattr(self, "_access_token") else None
        return create_path(self.root, access_token=access_token)

    def save(self, *args, **kwargs):
        """Save the storage record."""
        if hasattr(self, "_old_space") and hasattr(self, "_old_space_id"):
            if (
                self._old_space != self.space or self._old_space_id != self.space_id
            ):  # space_id is automatically handled by field tracker according to Claude
                update_storage_with_space(
                    storage_lnid=self.uid, space_lnid=self.space.uid
                )
        super().save(*args, **kwargs)
        return self

    def delete(self, permanent: bool | None = None) -> None:  # type: ignore
        # type ignore is there because we don't use a trash here unlike everywhere else
        """Delete the storage location.

        This errors in case the storage location is not empty.

        Unlike other `SQLRecord`-based registries, this does *not* move the storage record into the trash.

        Args:
            permanent: For consistency, `False` raises an error, as soft delete is impossible.
        """
        from .. import settings

        if permanent is False:
            raise ValueError(
                "Soft delete is not possible for Storage, "
                "use 'permanent=True' or 'permanent=None' for permanent deletion."
            )

        assert not self.artifacts.exists(), "Cannot delete storage holding artifacts."  # noqa: S101
        check_storage_is_empty(self.path)
        assert settings.storage.root_as_str != self.root, (  # noqa: S101
            "Cannot delete the current storage location, switch to another."
        )
        if setup_settings.user.handle != "anonymous":  # only attempt if authenticated
            storage_records = get_storage_records_for_instance(
                # only query those storage records on the hub that are managed by the current instance
                setup_settings.instance._id
            )
            for storage_record in storage_records:
                if storage_record["lnid"] == self.uid:
                    assert storage_record["is_default"] in {False, None}, (  # noqa: S101
                        "Cannot delete default storage of instance."
                    )
                    delete_storage_record(storage_record)
        ssettings = StorageSettings(self.root)
        if ssettings._mark_storage_root.exists():
            ssettings._mark_storage_root.unlink(
                missing_ok=True  # this is totally weird, but needed on Py3.11
            )
        super(SQLRecord, self).delete()



import bionty as bt
import lamindb as ln
import pytest


@pytest.fixture(scope="module")
def prepare_cell_type_registry():
    bt.CellType.filter().delete(permanent=True)
    records = [
        {
            "ontology_id": "CL:0000084",
            "name": "T cell",
            "synonyms": "T-cell|T-lymphocyte|T lymphocyte",
            "children": ["CL:0000798", "CL:0002420", "CL:0002419", "CL:0000789"],
        },
        {
            "ontology_id": "CL:0000236",
            "name": "B cell",
            "synonyms": "B-lymphocyte|B lymphocyte|B-cell",
            "children": ["CL:0009114", "CL:0001201"],
        },
        {
            "ontology_id": "CL:0000696",
            "name": "PP cell",
            "synonyms": "type F enteroendocrine cell",
            "children": ["CL:0002680"],
        },
        {
            "ontology_id": "CL:0002072",
            "name": "nodal myocyte",
            "synonyms": "P cell|myocytus nodalis|cardiac pacemaker cell",
            "children": ["CL:1000409", "CL:1000410"],
        },
    ]
    public_records = []
    for ref_record in records:
        record = bt.CellType.from_source(ontology_id=ref_record["ontology_id"])
        assert record.name == ref_record["name"]
        assert set(record.synonyms.split("|")) == set(ref_record["synonyms"].split("|"))
        public_records.append(record)
    ln.save(public_records)
    yield "prepared"
    bt.CellType.filter().delete(permanent=True)


def test_search_synonyms(prepare_cell_type_registry):
    result = bt.CellType.search("P cell").to_dataframe()
    assert set(result.name.iloc[:2]) == {"nodal myocyte", "PP cell"}


def test_search_limit(prepare_cell_type_registry):
    result = bt.CellType.search("P cell", limit=1).to_dataframe()
    assert len(result) == 1


def test_search_case_sensitive(prepare_cell_type_registry):
    result = bt.CellType.search("b cell", case_sensitive=False).to_dataframe()
    assert result.name.iloc[0] == "B cell"


def test_search_None():
    with pytest.raises(
        ValueError, match="Cannot search for None value! Please pass a valid string."
    ):
        bt.CellType.search(None)



# Generated by Django 5.2 on 2025-01-13 08:12

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


def migrate_refs_to_aux(apps, schema_editor):
    Reference = apps.get_model("lamindb", "Reference")

    for ref in Reference.objects.all():
        # Get existing _aux or initialize new dict
        aux_data = ref._aux or {}

        # Create or update the adhoc section
        aux_data["adhoc"] = {
            "preprint": ref.preprint,
            "public": ref.public,
            "journal": ref.journal,
        }

        ref._aux = aux_data
        ref.save(update_fields=["_aux"])


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0073_merge_ourprojects"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="ulabel",
            name="is_concept",
        ),
        migrations.AddField(
            model_name="ulabel",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="ulabel",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.ulabel",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="name",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=150
            ),
        ),
        migrations.AddField(
            model_name="feature",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="featureset",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="featureset",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.featureset",
            ),
        ),
        migrations.AddField(
            model_name="param",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.reference",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.param",
            ),
        ),
        migrations.RenameField(
            model_name="project",
            old_name="contributors",
            new_name="persons",
        ),
        migrations.AddField(
            model_name="project",
            name="parents",
            field=models.ManyToManyField(related_name="children", to="lamindb.project"),
        ),
        migrations.AddField(
            model_name="project",
            name="type",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=64, null=True
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="abbr",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=32, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="abbr",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=32, null=True
            ),
        ),
        migrations.RenameField(
            model_name="artifact",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="collection",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="feature",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="featureset",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="featurevalue",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="param",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="paramvalue",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="person",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="project",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="reference",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="run",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="storage",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="transform",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.RenameField(
            model_name="ulabel",
            old_name="aux",
            new_name="_aux",
        ),
        migrations.AddField(
            model_name="space",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_default="00000000",
                db_index=True,
                default="00000000",
                max_length=12,
                unique=True,
            ),
        ),
        # migrate references
        migrations.RunPython(
            migrate_refs_to_aux,
        ),
        # Then remove the old fields
        migrations.RemoveField(
            model_name="reference",
            name="preprint",
        ),
        migrations.RemoveField(
            model_name="reference",
            name="public",
        ),
        migrations.RemoveField(
            model_name="reference",
            name="journal",
        ),
        migrations.AlterField(
            model_name="artifact",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="_curator",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="featureset",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="paramvalue",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.RenameField(
            model_name="reference",
            old_name="published_at",
            new_name="date",
        ),
    ]
